/*
FRODO: a FRamework for Open/Distributed Optimization
Copyright (C) 2008, 2009, 2010, 2011  Thomas Leaute, Brammert Ottens & Radoslaw Szymanek

FRODO is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FRODO is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.


How to contact the authors: 
<http://lia.epfl.ch/>

EPFL / IC / IIF / LIA
Batiment IN 
Station 14 
CH - 1015 Lausanne 
Switzerland
*/

package ch.epfl.lia.frodo.algorithms;

import java.lang.reflect.Field;
import java.util.List;

import org.jdom.Document;
import org.jdom.Element;

import ch.epfl.lia.frodo.algorithms.varOrdering.election.VariableElection;
import ch.epfl.lia.frodo.communication.MessageListener;
import ch.epfl.lia.frodo.solutionSpaces.Addable;
import ch.epfl.lia.frodo.solutionSpaces.DCOPProblemInterface;

/** An abstract convenient class for solving DCOP instances
 * @author Thomas Leaute
 * @param <V> type used for variable values
 * @param <U> type used for utility values
 * @param <S> type used for the solution
 */
public abstract class AbstractDCOPsolver < V extends Addable<V>, U extends Addable<U>, S extends Solution<V, U> > 
	extends AbstractSolver<DCOPProblemInterface<V, U>, V, U, S> {

	/**
	 * Dummy constructor
	 */
	protected AbstractDCOPsolver() {
		super();
	}
	
	/** Constructor from an agent configuration file
	 * @param agentDescFile 	the agent configuration file
	 */
	protected AbstractDCOPsolver (String agentDescFile) {
		super (agentDescFile);
	}
	
	/** Constructor from an agent configuration file
	 * @param agentDescFile 	the agent configuration file
	 * @param useTCP 			Whether to use TCP pipes or shared memory pipes
	 * @warning Using TCP pipes automatically disables simulated time. 
	 */
	protected AbstractDCOPsolver (String agentDescFile, boolean useTCP) {
		super (agentDescFile, useTCP);
	}
	
	/** Constructor
	 * @param agentDesc 	a JDOM Document for the agent description
	 */
	protected AbstractDCOPsolver (Document agentDesc) {
		super(agentDesc);
	}
	
	/** Constructor
	 * @param agentDesc 	a JDOM Document for the agent description
	 * @param useTCP 		Whether to use TCP pipes or shared memory pipes
	 * @warning Using TCP pipes automatically disables simulated time. 
	 */
	protected AbstractDCOPsolver (Document agentDesc, boolean useTCP) {
		super (agentDesc, useTCP);
	}
	
	/** Constructor
	 * @param agentDesc 	The agent description
	 * @param parserClass 	The class of the parser to be used
	 */
	protected AbstractDCOPsolver (Document agentDesc, Class< ? extends XCSPparser<V, U> > parserClass) {
		super(agentDesc, parserClass);
	}
	
	/** Constructor
	 * @param agentDesc 	The agent description
	 * @param parserClass 	The class of the parser to be used
	 * @param useTCP 		Whether to use TCP pipes or shared memory pipes
	 * @warning Using TCP pipes automatically disables simulated time. 
	 */
	protected AbstractDCOPsolver (Document agentDesc, Class< ? extends XCSPparser<V, U> > parserClass, boolean useTCP) {
		super(agentDesc, parserClass, useTCP);
	}
	
	/** @see AbstractSolver#solve(org.jdom.Document, int, boolean, java.lang.Long, boolean) */
	@Override
	public S solve (Document problem, int nbrElectionRounds, boolean measureMsgs, Long timeout, boolean cleanAfterwards) {
		
		agentDesc.getRootElement().setAttribute("measureMsgs", Boolean.toString(measureMsgs));
		this.setNbrElectionRounds(nbrElectionRounds);
		return this.solve(problem, cleanAfterwards, timeout);
	}
	
	/** @see AbstractSolver#solve(ch.epfl.lia.frodo.solutionSpaces.ProblemInterface, int, boolean, java.lang.Long, boolean) */
	@Override
	public S solve (DCOPProblemInterface<V, U> problem, int nbrElectionRounds, boolean measureMsgs, Long timeout, boolean cleanAfterwards) {
		
		agentDesc.getRootElement().setAttribute("measureMsgs", Boolean.toString(measureMsgs));
		this.setNbrElectionRounds(nbrElectionRounds);
		return this.solve(problem, cleanAfterwards, timeout);
	}
	
	/** Sets the number of rounds of VariableElection
	 * @param nbrElectionRounds 	the number of rounds of VariableElection (must be greater than the diameter of the constraint graph)
	 */
	@SuppressWarnings("unchecked")
	protected void setNbrElectionRounds (int nbrElectionRounds) {
		for (Element module : (List<Element>) agentDesc.getRootElement().getChild("modules").getChildren()) 
			if (module.getAttributeValue("className").equals(VariableElection.class.getName())) 
				module.setAttribute("nbrSteps", Integer.toString(nbrElectionRounds));
	}
	
	/** @see AbstractSolver#solve(ch.epfl.lia.frodo.solutionSpaces.ProblemInterface, boolean, java.lang.Long) */
	@Override
	public S solve (DCOPProblemInterface<V, U> problem, boolean cleanAfterwards, Long timeout) {
		
		this.problem = problem;
		
		// Check whether it is necessary to instantiate the agent factory
		if (this.factory == null) {
			
			// Override message types if necessary
			this.overrideMsgTypes();

			// Instantiate the modules that listen for the solution
			this.solGatherers = this.getSolGatherers();

			// Solve the problem
			this.factory = new AgentFactory<V> (problem, agentDesc, solGatherers, timeout, this.useTCP);
			
		} else { // we already have an agent factory; just restart it
			
			for (StatsReporter module : this.solGatherers) 
				module.reset();
			this.factory.restart(problem);
		}
		
		if(factory.timedOut() || factory.outOfMemory()) {
			if (cleanAfterwards)
				this.clear();
			return null;
		}
		
		S solution = this.buildSolution();
		
		if (cleanAfterwards) 
			this.clear();

		return solution;
	}

	/** Overrides message types if necessary */
	@SuppressWarnings("unchecked")
	private void overrideMsgTypes() {

		Element modsElmt = agentDesc.getRootElement().getChild("modules");
		
		try {
			if (modsElmt != null) {
				for (Element moduleElmt : (List<Element>) modsElmt.getChildren()) {

					String className = moduleElmt.getAttributeValue("className");
					Class< MessageListener<String> > moduleClass = (Class< MessageListener<String> >) Class.forName(className);
					Element allMsgsElmt = moduleElmt.getChild("messages");
					if (allMsgsElmt != null) {
						for (Element msgElmt : (List<Element>) allMsgsElmt.getChildren()) {

							// Look up the new value for the message type
							String newType = msgElmt.getAttributeValue("value");
							String ownerClassName = msgElmt.getAttributeValue("ownerClass");
							if (ownerClassName != null) { // the attribute "value" actually refers to a field in a class
								Class<?> ownerClass = Class.forName(ownerClassName);
								try {
									Field field = ownerClass.getDeclaredField(newType);
									newType = (String) field.get(newType);
								} catch (NoSuchFieldException e) {
									System.err.println("Unable to read the value of the field " + ownerClass.getName() + "." + newType);
									e.printStackTrace();
								}
							}

							// Set the message type to its new value
							try {
								SingleQueueAgent.setMsgType(moduleClass, msgElmt.getAttributeValue("name"), newType);
							} catch (NoSuchFieldException e) {
								System.err.println("Unable to find the field " + moduleClass.getName() + "." + msgElmt.getAttributeValue("name"));
								e.printStackTrace();
							}
						}
					}
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
}
