/*
FRODO: a FRamework for Open/Distributed Optimization
Copyright (C) 2008, 2009, 2010, 2011  Thomas Leaute, Brammert Ottens & Radoslaw Szymanek

FRODO is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FRODO is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.


How to contact the authors: 
<http://lia.epfl.ch/>

EPFL / IC / IIF / LIA
Batiment IN 
Station 14 
CH - 1015 Lausanne 
Switzerland
 */

package ch.epfl.lia.frodo.algorithms.maxsum;

import java.lang.reflect.Array;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.jdom.Element;

import ch.epfl.lia.frodo.algorithms.AgentInterface;
import ch.epfl.lia.frodo.algorithms.StatsReporterWithConvergence;
import ch.epfl.lia.frodo.communication.Message;
import ch.epfl.lia.frodo.communication.MessageWith2Payloads;
import ch.epfl.lia.frodo.communication.Queue;
import ch.epfl.lia.frodo.solutionSpaces.Addable;
import ch.epfl.lia.frodo.solutionSpaces.AddableDelayed;
import ch.epfl.lia.frodo.solutionSpaces.DCOPProblemInterface;
import ch.epfl.lia.frodo.solutionSpaces.UtilitySolutionSpace;
import ch.epfl.lia.frodo.solutionSpaces.hypercube.Hypercube;

/** The Max-Sum Algorithm
 * 
 * Max-Sum Implementation mostly based on "Decentralised Coordination of Low-Power Embedded
 * Devices Using the Max-Sum Algorithm" by 
 * A. Farinelli, A. Rogers, A. Petcu, N. R. Jennings
 * 
 * Each function node is simulated by the agent owning the first variable in its scope. 
 * 
 * @author Thomas Leaute, remotely based on a preliminary contribution by Sokratis Vavilis and George Vouros (AI-Lab of the University of the Aegean)
 * 
 * @param <V> the type used for variable values
 * @param <U> the type used for utility values in stats gatherer mode
 */
public class MaxSum < V extends Addable<V>, U extends Addable<U> > implements StatsReporterWithConvergence<V> {
	
	/** The type of the message telling the agent to start the algorithm */
	public static String START_MSG_TYPE = AgentInterface.START_AGENT;

	/** The type of the messages received from function nodes of the graph */
	public static final String FUNCTION_MSG_TYPE = FunctionMsg.FUNCTION_MSG_TYPE;

	/** The type of the messages received from variable nodes of the graph */
	public static final String VARIABLE_MSG_TYPE = VariableMsg.VARIABLE_MSG_TYPE;

	/** The type of the messages containing the assignment history */
	private static final String CONV_STATS_MSG_TYPE = "MaxSumConvStatsMsg";

	/** The type of the messages containing the final assignment to a variable */
	private static final String SOLUTION_MSG_TYPE = "MaxSumSolMsg";
	
	/** Information about an internal variable */
	private class VarInfo {
		
		/** The name of this variable */
		private final String varName;
		
		/** The domain of the variable */
		private final V[] dom;
		
		/** The functions this variable is involved in */
		private final ArrayList<FunctionInfo> functions = new ArrayList<FunctionInfo> ();
		
		/** The last marginal utility received from each function node */     
		private HashMap< String, UtilitySolutionSpace<V, U> > lastMsgsIn = new HashMap< String, UtilitySolutionSpace<V, U> > ();
		
		/** The remaining number of cycles for this node */
		private int nbrCycles;
		
		/** The current optimal value for this variable */
		private V optVal;
		
		/** Constructor
		 * @param varName 		The name of this variable
		 */
		private VarInfo (String varName) {
			this.varName = varName;
			this.dom = problem.getDomain(varName);
			this.nbrCycles = maxNbrCycles;
			
			// Pick the first possible value for the variable
			this.optVal = this.dom[0];
		}
		
		/** Adds a function in which this variable is involved
		 * @param functionInfo 	the info about the function
		 */
		private void addFunction (FunctionInfo functionInfo) {
			
			this.functions.add(functionInfo);
			
			// Initialize the last messages received from and sent to this function node
			this.lastMsgsIn.put(functionInfo.name, zeroSpace(this.varName, this.dom));
		}
	}
	
	/** Returns a single-variable space full of zeros
	 * @param varName 	the name of the variable
	 * @param dom 		the variable domain
	 * @return the space
	 */
	@SuppressWarnings("unchecked")
	private Hypercube<V, U> zeroSpace (String varName, V[] dom) {
		
		U[] utils = (U[]) Array.newInstance(this.zero.getClass(), dom.length);
		Arrays.fill(utils, this.zero);
		V[][] doms = (V[][]) Array.newInstance(dom.getClass(), 1);
		doms[0] = dom;
		
		return new Hypercube<V, U> (new String[] { varName }, doms, utils, infeasibleUtil);
	}
	
	/** For each internal variable, its VarInfo */
	private HashMap<String, VarInfo> varInfos;
	
	/** Information about each neighboring function node */
	private class FunctionInfo {
		
		/** The name of the constraint */
		private final String name;
		
		/** The constraint */
		private final UtilitySolutionSpace<V, U> space;
		
		/** The agent responsible for simulating this function node */
		private final String agent;
		
		/** The last marginal utility received from each variable node */     
		private HashMap< String, UtilitySolutionSpace<V, U> > lastMsgsIn = new HashMap< String, UtilitySolutionSpace<V, U> > ();
		
		/** The remaining number of cycles for this node */
		private int nbrCycles;
		
		/** Constructor
		 * @param name 		The name of the constraint
		 * @param space 	The constraint
		 * @param agent 	The agent responsible for simulating this function node
		 */
		private FunctionInfo (String name, UtilitySolutionSpace<V, U> space, String agent) {
			this.name = name;
			this.agent = agent;
			this.nbrCycles = maxNbrCycles;
			this.space = space;
			
			for (String var : space.getVariables()) 
				this.addVariable(var, problem.getDomain(var));
		}
		
		/** Adds a variable to this FunctionInfo
		 * @param varName 	the variable name
		 * @param dom 		the variable domain
		 */
		private void addVariable (String varName, V[] dom) {
			
			// Initialize the last messages received from and sent to this variable node
			this.lastMsgsIn.put(varName, zeroSpace(varName, dom));
		}
	}
	
	/** For each constraint in the agent's subproblem, its FunctionInfo */
	private HashMap<String, FunctionInfo> functionInfos;

	/** The algorithm will terminate when ALL function nodes have gone through that many cycles */
	private final int maxNbrCycles;

	/** This module's queue */
	private Queue queue;

	/** The problem */
	private DCOPProblemInterface<V, U> problem;
	
	/** Whether the stats gatherer should display the solution found */
	private boolean silent = true;

	/** Whether the module has already started the algorithm */
	private boolean started = false;

	/** The 0 cost */
	private U zero;

	/** The solution */
	private HashMap<String, V> solution;

	/** The optimal cost */
	private U optCost;
	
	/** Whether to maximize utility or minimize cost */
	private final boolean maximize;

	/** The infeasible utility */
	private final U infeasibleUtil;
	
	/** Whether to rescale the messages sent by variable nodes */
	private final boolean rescale;

	/** Whether the listener should record the assignment history or not */
	private final boolean convergence;

	/** For each variable its assignment history */
	private final HashMap< String, ArrayList< CurrentAssignment<V> > > assignmentHistoriesMap;
	
	/** The owner of each variable */
	private Map<String, String> owners;

	/** Constructor
	 * @param problem       this agent's problem
	 * @param parameters    the parameters for this module
	 */
	public MaxSum (DCOPProblemInterface<V, U> problem, Element parameters) {
		
		this.problem = (DCOPProblemInterface<V, U>) problem;
		this.maximize = problem.maximize();
		this.infeasibleUtil = (this.maximize ? problem.getMinInfUtility() : problem.getPlusInfUtility());

		String convergence = parameters.getAttributeValue("convergence");
		if(convergence != null)
			this.convergence = Boolean.parseBoolean(convergence);
		else
			this.convergence = false;
		this.assignmentHistoriesMap = (this.convergence ? new HashMap< String, ArrayList< CurrentAssignment<V> > >() : null);

		String nbrCycles = parameters.getAttributeValue("nbrCycles");
		if(nbrCycles == null)
			this.maxNbrCycles = 200;
		else
			this.maxNbrCycles = Integer.parseInt(nbrCycles);
		
		this.rescale = Boolean.parseBoolean(parameters.getAttributeValue("rescale"));
	}

	/** The constructor called in "statistics gatherer" mode
	 * @param parameters    the description of what statistics should be reported (currently unused)
	 * @param problem       the overall problem
	 */
	public MaxSum (Element parameters, DCOPProblemInterface<V, U> problem)  {
		this.problem = problem;

		this.solution = new HashMap<String, V> ();
		this.optCost = problem.getZeroUtility();
		this.started = true;
		this.maximize = this.problem.maximize();
		this.infeasibleUtil = (this.maximize ? problem.getMinInfUtility() : problem.getPlusInfUtility());
		this.maxNbrCycles = 0;
		this.rescale = false;
		this.convergence = false;
		this.assignmentHistoriesMap = new HashMap< String, ArrayList< CurrentAssignment<V> > > ();

		// Display the factor graph if required
		if (parameters != null && Boolean.parseBoolean(parameters.getAttributeValue("reportStats"))) {
			this.silent = false;
			
			String dotRendererClass = parameters.getAttributeValue("DOTrenderer");
			if(dotRendererClass == null || dotRendererClass.equals("")) {
				System.out.println("Factor graph:");
				System.out.println(factorGraphToDOT());
			}
			else {
				try {
					Class.forName(dotRendererClass).getConstructor(String.class, String.class).newInstance("Factor graph", factorGraphToDOT());
				} 
				catch(Exception e) {
					System.out.println("Could not instantiate given DOT renderer class: " + dotRendererClass);
					e.printStackTrace();
				}
			}
		}
	}

	/** @return a DOT-formatted representation of the factor graph */
	private String factorGraphToDOT() {
		
		List< ? extends UtilitySolutionSpace<V, U> > allSpaces = this.problem.getSolutionSpaces();
		
		StringBuilder out = new StringBuilder ("graph {\n");
		
		// Print the agents, with their respective variable nodes and function nodes
		for (String agent : this.problem.getAgents()) {
			
			out.append("\tsubgraph cluster_" + agent + " {\n");
			out.append("\t\tlabel = " + agent + ";\n");
			
			// Add an invisible variable if the agent owns no variable, so that it is still displayed
			Set<String> myVars = this.problem.getVariables(agent);
			if (myVars.isEmpty()) 
				out.append("\t\t" + new Object().hashCode() + " [shape=\"none\", label=\"\"];\n");
			
			// Print the agent's variable nodes
			for (String var : myVars) 
				out.append("\t\t" + var + " [shape = \"circle\"];\n");
			out.append("\n");
					
			// Go through all constraints to find the ones this agent is responsible for 
			for (UtilitySolutionSpace<V, U> space : allSpaces) {
				if (agent.equals(this.problem.getOwner(space.getVariable(0)))) {
					out.append("\t\t" + space.getName() + " [shape = \"square\"];\n");
				}
			}

			out.append("\t}\n");
		}
		out.append("\n");
		
		// Print the edges between variable nodes and function nodes
		for (UtilitySolutionSpace<V, U> space : allSpaces) {
			for (String var : space.getVariables()) 
				out.append("\t" + space.getName() + "--" + var + ";\n");
		}
		out.append("\n");
				
		out.append("}\n");
		
		return out.toString();
	}

	/** Parses the problem */
	private void start () {
		
		this.started = true;
		
		// Terminate immediately if I own no variable
		Set<String> myVars = this.problem.getMyVars();
		if (myVars.isEmpty()) {
			this.queue.sendMessageToSelf(new Message (AgentInterface.AGENT_FINISHED));
			return;
		}

		this.zero = this.problem.getZeroUtility();
		
		this.varInfos = new HashMap<String, VarInfo> (myVars.size());
		for (String myVar : myVars) 
			this.varInfos.put(myVar, new VarInfo (myVar));
		this.owners = this.problem.getOwners();
				
		// Go through all constraints in this agent's subproblem
		this.functionInfos = new HashMap<String, FunctionInfo> ();
		for (UtilitySolutionSpace<V, U> space : this.problem.getSolutionSpaces()) {
			
			FunctionInfo info = new FunctionInfo (space.getName(), (UtilitySolutionSpace<V, U>) space, this.problem.getOwner(space.getVariable(0)));
			this.functionInfos.put(info.name, info);
			
			// Check which of my variables is involved in this function
			for (String var : space.getVariables()) {
				VarInfo varInfo = this.varInfos.get(var);
				if (varInfo != null) // internal variable
					varInfo.addFunction(info);
			}
		}
		
		// Start the algorithm by having each of my variable nodes send a fake message to each of its neighboring function nodes
		for (VarInfo varInfo : this.varInfos.values()) {
			
			// Check if this variable is unconstrained
			if (varInfo.functions.isEmpty()) {
				
				this.queue.sendMessage(AgentInterface.STATS_MONITOR, new MessageWith2Payloads<String, V> (SOLUTION_MSG_TYPE, varInfo.varName, varInfo.optVal));

				if(convergence) {
					ArrayList< CurrentAssignment<V> > history = this.assignmentHistoriesMap.get(varInfo.varName);
					history.add(new CurrentAssignment<V>(queue.getCurrentTime(), 0, varInfo.optVal));
					queue.sendMessage(AgentInterface.STATS_MONITOR, new StatsReporterWithConvergence.ConvStatMessage<V>(CONV_STATS_MSG_TYPE, varInfo.varName, history));
				}
				
				varInfo.nbrCycles = 0;
				this.checkForTermination();
			}
				
			else { // constrained variable
				for (FunctionInfo functionInfo : varInfo.functions) {
					UtilitySolutionSpace<V, U> space = zeroSpace (varInfo.varName, varInfo.dom);
					space.setName("start");
					this.queue.sendMessage(functionInfo.agent, new VariableMsg<V, U> (functionInfo.name, space));
				}
			}
		}
	}

	/** @see StatsReporterWithConvergence#reset() */
	public void reset() {
		/// @todo To be implemented
	}

	/** @see StatsReporterWithConvergence#getMsgTypes() */
	public Collection<String> getMsgTypes() {
		ArrayList<String> types = new ArrayList<String> (4);
		types.add(START_MSG_TYPE);
		types.add(FUNCTION_MSG_TYPE);
		types.add(VARIABLE_MSG_TYPE);
		types.add(AgentInterface.ALL_AGENTS_IDLE);
		return types;
	}

	/** @see StatsReporterWithConvergence#getStatsFromQueue(Queue) */
	public void getStatsFromQueue(Queue queue) {
		queue.addIncomingMessagePolicy(SOLUTION_MSG_TYPE, this);
		queue.addIncomingMessagePolicy(CONV_STATS_MSG_TYPE, this);
	}

	/** @see StatsReporterWithConvergence#setSilent(boolean) */
	public void setSilent(boolean silent) {
		this.silent = silent;
	}

	/** @see StatsReporterWithConvergence#setQueue(Queue) */
	public void setQueue(Queue queue) {
		this.queue = queue;
	}

	/** @see StatsReporterWithConvergence#notifyIn(Message) */
	@SuppressWarnings("unchecked")
	public void notifyIn(Message msg) {

		String msgType = msg.getType();

		if (msgType.equals(CONV_STATS_MSG_TYPE)) { // in stats gatherer mode, the message sent by a variable containing the assignment history
			
			StatsReporterWithConvergence.ConvStatMessage<V> msgCast = (StatsReporterWithConvergence.ConvStatMessage<V>)msg;
			assignmentHistoriesMap.put(msgCast.getVar(), msgCast.getAssignmentHistory());

			return;
			
		} else if (msgType.equals(SOLUTION_MSG_TYPE)) { // in stats gatherer mode, the final assignment to a variable
			
			MessageWith2Payloads<String, V> msgCast = (MessageWith2Payloads<String, V>) msg;
			String var = msgCast.getPayload1();
			V val = msgCast.getPayload2();
			
			this.solution.put(var, val);
			
			if (! this.silent) 
				System.out.println("var `" + var + "' = " + val);
			
			if (this.solution.size() == this.problem.getNbrVars()) {
				this.optCost = this.problem.getUtility(this.solution).getUtility(0);
				if (! this.silent) 
					System.out.println((this.maximize ? "Utility" : "Cost") + " of solution found: " + this.optCost);
			}
			
			return;
		}

		if (! this.started) 
			this.start();
		
//		System.out.println(msg);
		
		if (this.varInfos == null) // the agent has already terminated
			return;

		if (msgType.equals(FUNCTION_MSG_TYPE)) { // a message sent by a function node
			
			// Retrieve the information from the message
			FunctionMsg<V, U> msgCast = (FunctionMsg<V, U>) msg;
			UtilitySolutionSpace<V, U> marginalUtil = msgCast.getMarginalUtil();
			assert marginalUtil.getNumberOfVariables() == 1 : "Multi-variable marginal utility: " + marginalUtil;
			String var = marginalUtil.getVariable(0);
			VarInfo varInfo = this.varInfos.get(var);
			String functionNode = msgCast.getFunctionNode();
			
			if (--varInfo.nbrCycles == 0) {
				assert varInfo.optVal != null;
				this.queue.sendMessage(AgentInterface.STATS_MONITOR, new MessageWith2Payloads<String, V> (SOLUTION_MSG_TYPE, varInfo.varName, varInfo.optVal));
				if(convergence)
					queue.sendMessage(AgentInterface.STATS_MONITOR, new StatsReporterWithConvergence.ConvStatMessage<V>(CONV_STATS_MSG_TYPE, var, assignmentHistoriesMap.get(varInfo.varName)));
				
				this.checkForTermination();
				
			} else if (varInfo.nbrCycles < 0) // stop sending messages once I have exhausted my cycles
				return;
			
			// If this message hasn't changed since the last message received from this function node, re-send the same messages 
			if (marginalUtil.equivalent(varInfo.lastMsgsIn.get(functionNode))) 
				return;

			// Store the received message
			varInfo.lastMsgsIn.put(functionNode, marginalUtil);
			
			// Compute the new optimal assignment to the destination variable, as the argmax of the join of the marginal utilities received from all function nodes
			int newOptIndex = 0;
			U newOpt = this.infeasibleUtil;
			final int domSize = (int) marginalUtil.getNumberOfSolutions();
			for (int i = 0; i < domSize; i++) { // for each possible assignment to my variable
				
				AddableDelayed<U> sumDelayed = this.zero.addDelayed();
				for (UtilitySolutionSpace<V, U> space : varInfo.lastMsgsIn.values()) 
					sumDelayed.addDelayed(space.getUtility(i));
				U sum = sumDelayed.resolve();
				
				if (this.maximize ? sum.compareTo(newOpt) >= 0 : sum.compareTo(newOpt) <= 0) {
					newOpt = sum;
					newOptIndex = i;
				}
			}
			V newOptVal = varInfo.dom[newOptIndex];
			
			// Report the new optimal assignment if it has changed
			if (! newOptVal.equals(varInfo.optVal)) {
				varInfo.optVal = newOptVal;
				
//				System.out.println("var `" + varInfo.varName + "' gets assigned " + varInfo.optVal);
				
				if (this.convergence) 
					assignmentHistoriesMap.get(var).add(new CurrentAssignment<V>(queue.getCurrentTime(), 0, newOptVal));
			}
			
			// Compute and send a new message to each neighboring function node
			for (FunctionInfo functionInfo : varInfo.functions) {
				
				// Join all last marginal utilities received from all neighboring function nodes except the current one
				marginalUtil = this.zeroSpace(varInfo.varName, varInfo.dom);
				AddableDelayed<U> scalarDelayed = this.zero.addDelayed();
				for (Map.Entry< String, UtilitySolutionSpace<V, U> > entry : varInfo.lastMsgsIn.entrySet()) {
					if (! functionInfo.name.equals(entry.getKey())) {
						UtilitySolutionSpace<V, U> space = entry.getValue();
						for (int i = 0; i < domSize; i++) {
							U util = space.getUtility(i);
							marginalUtil.setUtility(i, marginalUtil.getUtility(i).add(util));
							if (this.rescale) 
								scalarDelayed.addDelayed(util);
						}
					}
				}
						
				if (this.rescale) { // rescale the join such that its utilities sum up to zero
					
					U scalar = scalarDelayed.resolve();
					scalar = scalar.divide(scalar.fromString(Integer.toString(domSize)));
					for (int i = 0; i < domSize; i++) 
						marginalUtil.setUtility(i, marginalUtil.getUtility(i).subtract(scalar));
				}
				
				// Send the message
				this.queue.sendMessage(functionInfo.agent, new VariableMsg<V, U> (functionInfo.name, marginalUtil));
			}
			
			
		} else if (msgType.equals(VARIABLE_MSG_TYPE)) { // a message sent by a variable node
			
			// Retrieve the information from the message
			VariableMsg<V, U> msgCast = (VariableMsg<V, U>) msg;
			FunctionInfo functionInfo = this.functionInfos.get(msgCast.getFunctionNode());
			UtilitySolutionSpace<V, U> marginalUtil = msgCast.getMarginalUtil();
			assert marginalUtil.getNumberOfVariables() == 1 : "Multi-variable marginal utility: " + marginalUtil;
			String senderVar = marginalUtil.getVariable(0);
			
			if (! "start".equals(marginalUtil.getName())) { // not the foo message sent at startup
				
				// If this message hasn't changed since the last message received from this function node, re-send the same messages
				functionInfo.nbrCycles--;
				if (marginalUtil.equivalent(functionInfo.lastMsgsIn.get(senderVar))) 
					return;
			}
			
			// Store the received message
			functionInfo.lastMsgsIn.put(senderVar, marginalUtil);
			
			// If I have exhausted all my cycles, only respond to variable nodes
			ArrayList<String> destinations = new ArrayList<String> (functionInfo.lastMsgsIn.keySet());
			if (--functionInfo.nbrCycles <= 0) 
				destinations.retainAll(Arrays.asList(senderVar));
			
			// Compute and send a message to each neighboring variable node
			final int nbrOtherSpaces = functionInfo.lastMsgsIn.size() - 1;
			for (String var : destinations) {
				
				marginalUtil = functionInfo.space;
				
				// Join with the last marginal utilities received from all variables except the destination variable
				if (nbrOtherSpaces > 0) {
					UtilitySolutionSpace<V, U>[] otherSpaces = new UtilitySolutionSpace [nbrOtherSpaces];
					int i = 0;
					for (Map.Entry< String, UtilitySolutionSpace<V, U> > entry : functionInfo.lastMsgsIn.entrySet()) 
						if (! var.equals(entry.getKey())) 
							otherSpaces[i++] = entry.getValue();
					marginalUtil = marginalUtil.join(otherSpaces);
				}
				
				// Blindly project all variables except the destination one
				String[] vars = new String [marginalUtil.getNumberOfVariables() - 1];
				int i = 0;
				for (String otherVar : marginalUtil.getVariables()) 
					if (! otherVar.equals(var)) 
						vars[i++] = otherVar;
				marginalUtil = marginalUtil.blindProject(vars, this.maximize);

				// Send the message
				this.queue.sendMessage(this.owners.get(var), new FunctionMsg<V, U> (functionInfo.name, marginalUtil.resolve()));
			}
			
		} else if (msgType.equals(AgentInterface.ALL_AGENTS_IDLE)) {
			
			// Send the stats
			for (VarInfo varInfo : this.varInfos.values()) {
				if (! varInfo.functions.isEmpty() && varInfo.nbrCycles > 0) { // unconstrained variables and variables with exhausted cycles have already been terminated
					
					assert varInfo.optVal != null;
					this.queue.sendMessage(AgentInterface.STATS_MONITOR, new MessageWith2Payloads<String, V> (SOLUTION_MSG_TYPE, varInfo.varName, varInfo.optVal));
					if(convergence)
						queue.sendMessage(AgentInterface.STATS_MONITOR, new StatsReporterWithConvergence.ConvStatMessage<V>(CONV_STATS_MSG_TYPE, varInfo.varName, assignmentHistoriesMap.get(varInfo.varName)));
				}
			}
			
			this.queue.sendMessageToSelf(new Message (AgentInterface.AGENT_FINISHED));
			this.varInfos = null;
		}
	}

	/** Checks if all my variable nodes have finished */
	private void checkForTermination() {
		
		for (VarInfo info : this.varInfos.values()) 
			if (info.nbrCycles > 0) 
				return;
				
		this.queue.sendMessageToSelf(new Message (AgentInterface.AGENT_FINISHED));
		this.varInfos = null;
	}

	/** @return for each variable in the problem, its chosen value */
	public HashMap<String, V> getOptAssignments () {
		return this.solution;
	}

	/** @return the cost of the optimal solution found */
	public U getOptCost () {
		return this.optCost;
	}

	/** @see StatsReporterWithConvergence#getAssignmentHistories() */
	public HashMap< String, ArrayList< CurrentAssignment<V> > > getAssignmentHistories() {
		return this.assignmentHistoriesMap;
	}

	/** @see StatsReporterWithConvergence#getCurrentSolution() */
	public Map<String, V> getCurrentSolution() {
		/// @todo Auto-generated method stub
		assert false: "Not Implemented";
		return null;
	}

}