/*
FRODO: a FRamework for Open/Distributed Optimization
Copyright (C) 2008, 2009, 2010, 2011  Thomas Leaute, Brammert Ottens & Radoslaw Szymanek

FRODO is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FRODO is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.


How to contact the authors: 
<http://lia.epfl.ch/>

EPFL / IC / IIF / LIA
Batiment IN 
Station 14 
CH - 1015 Lausanne 
Switzerland
*/

package ch.epfl.lia.frodo.algorithms.subsolvers;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.jdom.Document;
import org.jdom.Element;

import ch.epfl.lia.frodo.algorithms.XCSPparser;
import ch.epfl.lia.frodo.solutionSpaces.Addable;
import ch.epfl.lia.frodo.solutionSpaces.AddableReal;
import ch.epfl.lia.frodo.solutionSpaces.BasicUtilitySolutionSpace;
import ch.epfl.lia.frodo.solutionSpaces.ProblemInterface;
import ch.epfl.lia.frodo.solutionSpaces.SolutionSpace;
import ch.epfl.lia.frodo.solutionSpaces.UtilitySolutionSpace;
import ch.epfl.lia.frodo.solutionSpaces.hypercube.HyperCubeIterBestFirst;
import ch.epfl.lia.frodo.solutionSpaces.hypercube.Hypercube;

/** This local solver hides backyard variables from the caller
 * 
 * Backyard variables are internal variables only constrained with other internal variables. 
 * This local solver makes the problem look as if backyard variables did not exist: 
 * they are properly handled internally, but do not participate directly in the DCOP; 
 * in particular, they do not appear in the DFS. 
 * 
 * @note There is one exception to the definition of backyard variables: if there exists one graph component 
 * consisting only of backyard variables, then one of the variables should be promoted to the facade so that 
 * the component does not get completely ignored by the DCOP algorithm. The first variable in lexicographic
 * order is selected. 
 * 
 * @author Thomas Leaute
 * @param <U> the type used for variable values
 * @param <V> the type used for utility values
 */
public class SimpleSubsolver < V extends Addable<V>, U extends Addable<U> > 
extends XCSPparser<V, U> {
	
	/** Used for serialization */
	private static final long serialVersionUID = -3162166845945946417L;

	/** The set of backyard variables, i.e. internal variables only connected to internal variables */
	private Set<String> backyard;
	
	/** The list of connected components in the constraint graph from which facade and external variables have been removed */
	private ArrayList< HashSet<String> > backyardComponents;

	/** The list of CompoundSpace scopes (one constraint per connected backyard component) */
	private ArrayList< HashSet<String> > facadeConstraints;
	
	/** For each facade variable, its non-backyard neighbors */
	private Map< String, HashSet<String> > facadeNeighborhoods;
	
	/** Constructor from a JDOM Document in XCSP format
	 * @param doc 	the JDOM Document in XCSP format
	 */
	public SimpleSubsolver(Document doc) {
		super(doc);
	}

	/** Constructor from a JDOM Document in XCSP format
	 * @param doc 			the JDOM Document in XCSP format
	 * @param countNCCCs 	Whether to count constraint checks
	 */
	public SimpleSubsolver(Document doc, Boolean countNCCCs) {
		super(doc, countNCCCs);
	}

	/** Constructor
	 * @param doc 		the JDOM Document in XCSP format
	 * @param params 	the parameters of the solver
	 */
	public SimpleSubsolver(Document doc, Element params) {
		super(doc, params);
	}

	/** Constructor from a JDOM root Element in XCSP format
	 * @param agentName 					the name of the agent owning the input subproblem
	 * @param root 							the JDOM root Element in XCSP format
	 * @param countNCCCs 					Whether to count constraint checks
	 * @param extendedRandNeighborhoods 	whether we want extended random neighborhoods
	 * @param spacesToIgnoreNcccs			list of spaces for which NCCCs should NOT be counted
	 * @param mpc 							Whether to behave in MPC mode
	 */
	protected SimpleSubsolver(String agentName, Element root, boolean countNCCCs, boolean extendedRandNeighborhoods, HashSet<String> spacesToIgnoreNcccs, boolean mpc) {
		super(agentName, root, countNCCCs, extendedRandNeighborhoods, spacesToIgnoreNcccs, mpc);
	}

	/** @see XCSPparser#newInstance(String, Element) */
	@Override
	protected SimpleSubsolver<V, U> newInstance (String agent, Element instance) {
		return new SimpleSubsolver<V, U> (agent, instance, this.countNCCCs, super.extendedRandNeighborhoods, this.spacesToIgnoreNcccs, super.mpc);
	}
	
	/** Computes the backyard of the input agent and the facade neighborhoods 
	 * 
	 * Imagine that you remove all facade and external variables from the problem. 
	 * You end up with (possibly several) disconnected components, composed of only backyard variables. 
	 * Each such component defines a CompoundSpace over the facade and external variables at its frontier. 
	 * The facade neighborhoods are then the neighborhoods defined by the CombinedSpaces. 
	 * 
	 * @param agent 	the name of the agent; if \c null, proceed for all agents
	 */
	private void computeBackyard (String agent) {
		
		if (this.backyard == null) {
			this.backyard = new HashSet<String> ();
			this.backyardComponents = new ArrayList< HashSet<String> > ();
			this.facadeConstraints = new ArrayList< HashSet<String> > ();
			this.facadeNeighborhoods = new HashMap< String, HashSet<String> > ();
		}
		
		// If agent == null, compute all agents' backyards
		if (agent == null) {
			for (String agent2 : super.getAgents()) 
				this.computeBackyard(agent2);
			return;
		}
		
		// Compute the "normal" neighborhoods
		Map< String, HashSet<String> > neighborhoods = new HashMap< String, HashSet<String> > ();
		for (String var : (agent == null ? this.getVariables() : this.getVariables(agent))) {
			
			// Get the neighbors of this variable
			HashSet<String> neighbors = super.getNeighborVars(var, true);
			neighborhoods.put(var, neighbors);
		}

		// Go through the list of neighbors of each internal variable
		HashSet<String> remaining = new HashSet<String> ();
		ext: for (Map.Entry< String, HashSet<String> > entry : neighborhoods.entrySet()) {
			String var = entry.getKey();
			
			// If any of the neighbors of this internal variable belongs to another agent, skip this variable
			for (String neighbor : entry.getValue()) {
				String owner = super.getOwner(neighbor);
				if (owner != null && ! agent.equals(owner)) 
					continue ext;
			}
			
			// All neighbors of this internal variable are internal; this is therefore a backyard variable
			this.backyard.add(var);
			remaining.add(var);
		}
		
		// Loop while there are still backyard variables remaining
		while(! remaining.isEmpty()) {
			
			HashSet<String> backComp = new HashSet<String> ();
			this.backyardComponents.add(backComp);
			HashSet<String> facadeConst = new HashSet<String> ();
			this.facadeConstraints.add(facadeConst);
			
			// Traverse the constraint graph in DFS order 
			Iterator<String> iter = remaining.iterator();
			String var = iter.next();
			iter.remove();
			backComp.add(var);
			String promotedVar = traverse (neighborhoods, var, remaining, backComp, facadeConst);
			
			// Promote the selected variable to the facade if necessary
			if (promotedVar != null) {
				this.backyard.remove(promotedVar);
				backComp.remove(promotedVar);
				facadeConst.add(promotedVar);
				this.facadeNeighborhoods.put(promotedVar, new HashSet<String> ());
				continue;
			}
			
			// Record the facade neighborhoods
			for (String var2 : facadeConst) {
				
				// Skip this variable if it is a random variable
				if (this.getOwner(var2) == null) 
					continue;

				HashSet<String> neighbors = this.facadeNeighborhoods.get(var2);
				if (neighbors == null) {
					neighbors = new HashSet<String> (facadeConst);
					this.facadeNeighborhoods.put(var2, neighbors);
				} else 
					neighbors.addAll(facadeConst);
				neighbors.remove(var2);
			}
		}
		
		// Also compute facade neighborhoods resulting from constraints not involving backyard variables
		for (Map.Entry< String, HashSet<String> > entry : neighborhoods.entrySet()) {
			
			// Skip backyard variables and random variables
			String var = entry.getKey();
			if (this.backyard.contains(var) || this.getOwner(var) == null) 
				continue;
			
			HashSet<String> neighbors = this.facadeNeighborhoods.get(var);
			if (neighbors == null) {
				neighbors = new HashSet<String> ();
				this.facadeNeighborhoods.put(var, neighbors);
			}
			
			for (String neigh : entry.getValue()) 
				if (! this.backyard.contains(neigh)) 
					neighbors.add(neigh);
		}
	}
	
	/** Traverses the constraint graph to compute the facade neighborhoods
	 * @param neighborhoods 	neighborhoods involving all variables
	 * @param var 				the current backyard variable
	 * @param remaining 		remaining backyard variables
	 * @param backComp 			backyard variables belonging to the current component
	 * @param facadeConst 		facade and external variables at the frontier of the current component
	 * @return \c null if the component contains at least one facade variable, else the first backyard variable in lexicographic order
	 */
	private String traverse (Map< String, HashSet<String> > neighborhoods, String var, HashSet<String> remaining, HashSet<String> backComp, HashSet<String> facadeConst) {
		
		String promotedVar = var;
		
		// Go through the list of neighbors of the current variable
		for (String neigh : neighborhoods.get(var)) {
			
			// Check if neigh is a backyard variable
			if (this.backyard.contains(neigh)) {
				if (remaining.remove(neigh)) { // neigh has not been treated yet
					backComp.add(neigh);
					String proposedVar = traverse(neighborhoods, neigh, remaining, backComp, facadeConst);
					if (promotedVar != null && (proposedVar == null || promotedVar.compareTo(proposedVar) > 0)) 
						promotedVar = proposedVar;
				}
			} else { // not a backyard variable
				facadeConst.add(neigh);
				if (! this.isRandom(neigh)) 
					promotedVar = null;
			}
		}
		
		return promotedVar;
	}

	/** A very special space that supports backyard variables
	 * @param <U> the type used for variable values
	 * @param <V> the type used for utility values
	 * @todo Write unit tests. 
	 */
	static class CompoundSpace < V extends Addable<V>, U extends Addable<U> > 
	implements UtilitySolutionSpace<V, U> {
		
		/** Used for serialization */
		private static final long serialVersionUID = 3667280312695632424L;

		/** The facade variables */
		private String[] facade;
		
		/** The backyard variables */
		private HashSet<String> backyard;
		
		/** The spaces that make up this compound space */
		private ArrayList< UtilitySolutionSpace<V, U> > spaces;

		/** Whether we should maximize of minimize over the backyard variables */
		private boolean maximize;

		/** Constructor
		 * @param facade 	facade variables
		 * @param backyard 	backyard variables
		 * @param spaces 	spaces that make up this compound space
		 * @param maximize 	whether this is a maximization or minimization problem
		 */
		public CompoundSpace(Collection<String> facade, HashSet<String> backyard, ArrayList< UtilitySolutionSpace<V, U> > spaces, boolean maximize) {
			this.facade = facade.toArray(new String [facade.size()]);
			this.backyard = backyard;
			this.spaces = spaces;
			this.maximize = maximize;
		}
		
		/** @see Object#toString() */
		@Override
		public String toString () {
			
			StringBuilder builder = new StringBuilder ("CompoundSpace:\n\tfacade: ");
			
			// Print the facade
			for (String var : facade) 
				builder.append(var + " ");
			
			builder.append("\n\tbackyard: " + backyard);
			
			builder.append("\n\tspaces:\n" + spaces + "\n");
			
			return builder.toString();
		}
		
		/** @see BasicUtilitySolutionSpace#prettyPrint(java.io.Serializable) */
		public String prettyPrint (U unused) {
			return this.toString();
		}
		
		/** @return the spaces in this compound space */
		ArrayList< UtilitySolutionSpace<V, U> > getSpaces () {
			return this.spaces;
		}

		/** @see UtilitySolutionSpace#changeVariablesOrder(String[]) */
		public UtilitySolutionSpace<V, U> changeVariablesOrder (String[] variables_order) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}
		
		/** @see UtilitySolutionSpace#clone() */
		public UtilitySolutionSpace<V, U> clone () {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** @see UtilitySolutionSpace#compose(String[], BasicUtilitySolutionSpace) */
		public UtilitySolutionSpace<V, U> compose (String[] vars, BasicUtilitySolutionSpace< V, ArrayList<V> > substitution) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** @see UtilitySolutionSpace#consensus(String, Map, boolean) */
		public ProjOutput<V, U> consensus (String varOut, Map<String, UtilitySolutionSpace< V, AddableReal> > distributions, boolean maximum) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** @see UtilitySolutionSpace#consensusAllSols(String, Map, boolean) */
		public ProjOutput<V, U> consensusAllSols (String varOut, Map<String, UtilitySolutionSpace< V, AddableReal> > distributions, boolean maximum) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** @see UtilitySolutionSpace#expectation(java.util.Map) */
		public UtilitySolutionSpace<V, U> expectation (Map< String, UtilitySolutionSpace<V, U> > distributions) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** @see UtilitySolutionSpace#isIncludedIn(UtilitySolutionSpace) */
		public boolean isIncludedIn (UtilitySolutionSpace<V, U> space) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return false;
		}

		/** @see UtilitySolutionSpace#join(UtilitySolutionSpace, String[]) */
		public UtilitySolutionSpace<V, U> join (UtilitySolutionSpace<V, U> space, String[] total_variables) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** 
		 * @see UtilitySolutionSpace#join(UtilitySolutionSpace) 
		 * @warning Assumes that the input is not a CompoundSpace. 
		 */
		public UtilitySolutionSpace<V, U> join (UtilitySolutionSpace<V, U> space) {
			
			// Construct the new list of variables
			HashSet<String> newVars = new HashSet<String> ();
			for (String var : this.facade) 
				newVars.add(var);
			for (String var : space.getVariables()) 
				newVars.add(var);
			
			// Construct the new list of spaces
			ArrayList< UtilitySolutionSpace<V, U > > newSpaces = new ArrayList< UtilitySolutionSpace<V, U > > (this.spaces);
			HashSet<String> newBackyard = new HashSet<String> (this.backyard);
			if (space instanceof CompoundSpace<?, ?>) {
				CompoundSpace<V, U> spaceCast = (CompoundSpace<V, U>) space;
				newSpaces.addAll(spaceCast.spaces);
				newBackyard.addAll(spaceCast.backyard);
			} else {
				newSpaces.add(space);
			}
			
			return new CompoundSpace<V, U> (newVars, newBackyard, newSpaces, this.maximize);
		}

		/** @see UtilitySolutionSpace#joinMinNCCCs(UtilitySolutionSpace) */
		public UtilitySolutionSpace<V, U> joinMinNCCCs(UtilitySolutionSpace<V, U> space) {
			System.err.println("NCCCs make no sense when using a local subsolver");
			return null;
		}

		/** 
		 * @see UtilitySolutionSpace#join(UtilitySolutionSpace[]) 
		 * @warning Assumes that the inputs are not CompoundSpaces. 
		 */
		public UtilitySolutionSpace<V, U> join (UtilitySolutionSpace<V, U>[] spaces) {

			// Construct the new list of variables and spaces
			HashSet<String> newVars = new HashSet<String> ();
			HashSet<String> newBackyard = new HashSet<String> (this.backyard);
			for (String var : this.facade) 
				newVars.add(var);
			ArrayList< UtilitySolutionSpace<V, U > > newSpaces = new ArrayList< UtilitySolutionSpace<V, U > > (this.spaces);
			for (UtilitySolutionSpace<V, U> space : spaces) {
				for (String var : space.getVariables()) 
					newVars.add(var);

				if (space instanceof CompoundSpace<?, ?>) {
					CompoundSpace<V, U> spaceCast = (CompoundSpace<V, U>) space;
					newSpaces.addAll(spaceCast.spaces);
					newBackyard.addAll(spaceCast.backyard);
				} else {
					newSpaces.add(space);
				}
			}
			
			return new CompoundSpace<V, U> (newVars, newBackyard, newSpaces, this.maximize);
		}

		/** @see UtilitySolutionSpace#joinMinNCCCs(UtilitySolutionSpace[]) */
		public UtilitySolutionSpace<V, U> joinMinNCCCs(UtilitySolutionSpace<V, U>[] spaces) {
			System.err.println("NCCCs make no sense when using a local subsolver");
			return null;
		}
		
		/** @see UtilitySolutionSpace#multiply(UtilitySolutionSpace, String[]) */
		public UtilitySolutionSpace<V, U> multiply (UtilitySolutionSpace<V, U> space, String[] total_variables) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** @see UtilitySolutionSpace#multiply(UtilitySolutionSpace) */
		public UtilitySolutionSpace<V, U> multiply (UtilitySolutionSpace<V, U> space) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** @see UtilitySolutionSpace#multiply(UtilitySolutionSpace[]) */
		public UtilitySolutionSpace<V, U> multiply(UtilitySolutionSpace<V, U>[] spaces) {
			/// @todo Auto-generated method stub
			assert false : "Not yet implemented";
			return null;
		}
		
		/** @see UtilitySolutionSpace#project(String[], boolean) */
		public ProjOutput<V, U> project (String[] variables_names, boolean maximum) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** @see UtilitySolutionSpace#project(int, boolean) */
		public ProjOutput<V, U> project (int number_to_project, boolean maximum) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}
		
		/** @see UtilitySolutionSpace#min(java.lang.String) */
		public UtilitySolutionSpace<V, U> min (String variable) {
			return this.blindProject(variable, false);
		}

		/** @see UtilitySolutionSpace#max(java.lang.String) */
		public UtilitySolutionSpace<V, U> max (String variable) {
			return this.blindProject(variable, true);
		}

		/** 
		 * @see UtilitySolutionSpace#project(String, boolean) 
		 * @todo Instead of immediately projecting all backyard variables, choosing carefully the ones whose projection should be delayed would yield a smaller output. 
		 */
		@SuppressWarnings("unchecked")
		public ProjOutput<V, U> project (String variable_name, final boolean maximum) {
			
			// Build up the list of variables that must be projected out
			String[] varsOut = new String [this.backyard.size() + 1];
			int i = 0;
			varsOut[i++] = variable_name;
			for (String var : this.backyard) 
				varsOut[i++] = var;
			
			int nbrSpaces = this.spaces.size();
			switch (nbrSpaces) {

			case 1: // only one space
				return this.spaces.get(0).project(varsOut, maximum);
				
			case 2: // only two spaces
				return this.spaces.get(0).join(this.spaces.get(1)).project(varsOut, maximum);
				
			default: // more than two spaces
				
				/// @todo Instead of joining everything and projecting everything at once, use a DPOP-like approach to join and project in a smart order, variable by variable.  
				
				UtilitySolutionSpace<V, U>[] otherSpaces = new UtilitySolutionSpace [--nbrSpaces];
				for (i = 0; i < nbrSpaces; i++) 
					otherSpaces[i] = this.spaces.get(i);
				return this.spaces.get(nbrSpaces).join(otherSpaces).project(varsOut, maximum);
			}
		}

		/** @see UtilitySolutionSpace#blindProject(java.lang.String, boolean) */
		public UtilitySolutionSpace<V, U> blindProject(String varOut, boolean maximize) {
			/// @todo Auto-generated method stub
			assert false : "Not yet implemented";
			return null;
		}

		/** @see UtilitySolutionSpace#blindProject(java.lang.String[], boolean) */
		public UtilitySolutionSpace<V, U> blindProject(String[] varsOut, boolean maximize) {
			/// @todo Auto-generated method stub
			assert false : "Not yet implemented";
			return null;
		}

		/** @see UtilitySolutionSpace#blindProjectAll(boolean) */
		public U blindProjectAll(boolean maximize) {
			/// @todo Auto-generated method stub
			assert false : "Not yet implemented";
			return null;
		}
		
		/** @see UtilitySolutionSpace#sample(int) */
		public Map<V, Double> sample (int nbrSamples) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** @see UtilitySolutionSpace#split(Addable, boolean) */
		public UtilitySolutionSpace<V, U> split (U threshold, boolean maximum) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** Augments this BasicUtilitySolutionSpace with a new combination of variables values associated with a given utility
		 * @note If the corresponding assignment already exists in BasicUtilitySolutionSpace, 
		 * this method will only modify the corresponding utility value. 
		 * @param variables_values assignments to the variables
		 * @param utility_value    corresponding utility value
		 */
		public void augment (V[] variables_values, U utility_value) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			
		}
		
		/** @see Object#equals(Object) */
		@SuppressWarnings("unchecked")
		@Override
		public boolean equals (Object o) {
			
			if (o == this)
				return true;
			
			if (!(o instanceof CompoundSpace))
				return false;
			
			CompoundSpace<V, U> other = (CompoundSpace<V, U>) o;
			
			if (! this.backyard.equals(other.backyard)) 
				return false;
			
			// Check the facades
			if (this.facade.length != other.facade.length) 
				return false;
			ext: for (String var : this.facade) {
				for (String var2 : other.facade) {
					if (var.equals(var2)) 
						continue ext; 
				}
				return false; // var is not in other.facade
			}
			
			// Check the spaces
			if (this.spaces.size() != other.spaces.size()) 
				return false;
			ext: for (UtilitySolutionSpace<V, U> space : this.spaces) {
				for (UtilitySolutionSpace<V, U> space2 : other.spaces) {
					if (space.equals(space2)) 
						continue ext;
				}
				return false; // space is not in other.spaces
			}
			
			return true;
		}

		/** @see BasicUtilitySolutionSpace#equivalent(BasicUtilitySolutionSpace) */
		public boolean equivalent (BasicUtilitySolutionSpace<V, U> space) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return false;
		}

		/** @see BasicUtilitySolutionSpace#getDefaultUtility() */
		public U getDefaultUtility () {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** @see BasicUtilitySolutionSpace#getUtility(V[]) */
		public U getUtility (V[] variables_values) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** @see BasicUtilitySolutionSpace#getUtility(String[], V[]) */
		public U getUtility (String[] variables_names, V[] variables_values) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** @see BasicUtilitySolutionSpace#getUtility(java.util.Map) */
		public U getUtility(Map<String, V> assignments) {
			/// @todo Auto-generated method stub
			assert false : "Not implemented";
			return null;
		}

		/** @see BasicUtilitySolutionSpace#getUtility(long) */
		public U getUtility (long index) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** @see UtilitySolutionSpace#setProblem(ProblemInterface) */
		public void setProblem(ProblemInterface<V, U> problem) {
			for (UtilitySolutionSpace<V, U> space : this.spaces) 
				space.setProblem(problem);
		}

		/** @see BasicUtilitySolutionSpace#isIncludedIn(BasicUtilitySolutionSpace) */
		public boolean isIncludedIn (BasicUtilitySolutionSpace<V, U> space) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return false;
		}

		/** @see UtilitySolutionSpace#iterator() */
		public UtilitySolutionSpace.Iterator<V, U> iterator () {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** @see UtilitySolutionSpace#iterator(String[], V[][]) */
		public UtilitySolutionSpace.Iterator<V, U> iterator (String[] variables, V[][] domains) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** Sets the default utility value
		 * 
		 * The default utility value is the one that is used when augmenting the BasicUtilitySolutionSpace with a new solution
		 * without specifying the utility of this solution. 
		 * @param utility 	the new default utility
		 */
		public void setDefaultUtility (U utility) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			
		}

		/** @see BasicUtilitySolutionSpace#setInfeasibleUtility(java.io.Serializable) */
		public void setInfeasibleUtility(U utility) {
			for (UtilitySolutionSpace<V, U> space : this.spaces) 
				space.setInfeasibleUtility(utility);
		}

		/** Sets the utility value corresponding to a given assignment to variables
		 * @param variables_values 	values for the variables, in the same order as in the BasicUtilitySolutionSpace
		 * @param utility 			the new utility value
		 * @return specifies if setting the utility value was possible. \c false is returned
		 * if the BasicUtilitySolutionSpace does not allow the given assignments to the variables.
		 */
		public boolean setUtility (V[] variables_values, U utility) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return false;
		}

		/** @see BasicUtilitySolutionSpace#setUtility(long, java.io.Serializable) */
		public void setUtility (long index, U utility) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			
		}

		/** @see UtilitySolutionSpace#slice(String[], V[][]) */
		public UtilitySolutionSpace<V, U> slice (String[] variables_names, V[][] sub_domains) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** @see UtilitySolutionSpace#slice(String[], V[]) 
		 * @todo Inefficient implementation */
		public UtilitySolutionSpace<V, U> slice (String[] variables_names, V[] values) {
			
			// Slice each space and join everything together
			java.util.Iterator< UtilitySolutionSpace<V, U> > iter = this.spaces.iterator();
			UtilitySolutionSpace<V, U> join = iter.next().slice(variables_names, values);
			while (iter.hasNext()) 
				join = join.join(iter.next().slice(variables_names, values));
			
			// Project out all facade variables and return
			return join.blindProject(this.backyard.toArray(new String [this.backyard.size()]), this.maximize);
		}

		/** @see UtilitySolutionSpace#slice(String, V[]) */
		public UtilitySolutionSpace<V, U> slice (String var, V[] subDomain) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** @see UtilitySolutionSpace#slice(String, Addable) */
		public UtilitySolutionSpace<V, U> slice (String var, V val) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** @see UtilitySolutionSpace#slice(V[]) */
		public UtilitySolutionSpace<V, U> slice (V[] variables_values) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** @see SolutionSpace#augment(V[]) */
		public void augment (V[] variables_values) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			
		}

		/** @see SolutionSpace#getDomain(String) */
		public V[] getDomain (String variable) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** @see SolutionSpace#getDomain(int) */
		public V[] getDomain (int index) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** @see SolutionSpace#getDomain(String, int) */
		public V[] getDomain (String variable, int index) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** @see SolutionSpace#getDomains() */
		public V[][] getDomains () {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** @see SolutionSpace#setDomain(String, V[]) */
		public void setDomain(String var, V[] dom) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
		}

		/** @see SolutionSpace#getIndex(String) */
		public int getIndex (String variable) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return 0;
		}

		/** @see SolutionSpace#getNumberOfSolutions() */
		public long getNumberOfSolutions () {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return 0;
		}

		/** @see SolutionSpace#getNumberOfVariables() */
		public int getNumberOfVariables () {
			return this.facade.length;
		}

		/** @see SolutionSpace#getVariable(int) */
		public String getVariable (int index) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** @see SolutionSpace#getVariables() */
		public String[] getVariables () {
			return this.facade;
		}

		/** @see SolutionSpace#renameVariable(String, String) */
		public void renameVariable(String oldName, String newName) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
		}

		/** @see SolutionSpace#renameAllVars(String[]) */
		public SolutionSpace<V> renameAllVars(String[] newVarNames) {
			/// @todo Auto-generated method stub
			assert false : "Not yet implemented";
			return null;
		}
		
		/** @see SolutionSpace#join(SolutionSpace, String[]) */
		public SolutionSpace<V> join (SolutionSpace<V> space, String[] total_variables) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** @see SolutionSpace#join(SolutionSpace) */
		public SolutionSpace<V> join (SolutionSpace<V> space) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** @see SolutionSpace#join(SolutionSpace[], String[]) */
		public SolutionSpace<V> join (SolutionSpace<V>[] spaces, String[] total_variables_order) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** @see SolutionSpace#join(SolutionSpace[]) */
		public SolutionSpace<V> join (SolutionSpace<V>[] spaces) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** Always returns \c true
		 * @see SolutionSpace#knows(Class) 
		 */
		public boolean knows(Class<?> spaceClass) {
			return true;
		}

		/** @see SolutionSpace#getName() */
		public String getName() {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** @see SolutionSpace#setName(String) */
		public void setName(String name) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
		}

		/** @see SolutionSpace#getRelationName() */
		public String getRelationName() {
			/// @todo Auto-generated method stub
			assert false : "Not yet implemented";
			return null;
		}

		/** @see SolutionSpace#setRelationName(String) */
		public void setRelationName(String name) {
			/// @todo Auto-generated method stub
			assert false : "Not yet implemented";
			
		}
		
		/** 
		 * @see ch.epfl.lia.frodo.solutionSpaces.UtilitySolutionSpace#iteratorBestFirst(boolean)
		 */
		public HyperCubeIterBestFirst<V, U> iteratorBestFirst(boolean maximize) {
			/// @todo Auto-generated method stub
			assert false : "not implemented!";
			return null;
		}

		/** @see BasicUtilitySolutionSpace#getClassOfU() */
		public Class<U> getClassOfU() {
			/// @todo Auto-generated method stub
			assert false : "Not yet implemented";
			return null;
		}

		/**@see UtilitySolutionSpace#iterator(java.lang.String[]) */
		public UtilitySolutionSpace.Iterator<V, U> iterator(String[] order) {
			/// @todo Auto-generated method stub
			assert false : "Not yet implemented";
			return null;
		}

		/** @see UtilitySolutionSpace#projectAll(boolean) */
		public ProjOutput<V, U> projectAll(
				boolean maximum) {
			/// @todo Auto-generated method stub
			assert false : "Not yet implemented";
			return null;
		}

		/** @see UtilitySolutionSpace#projectAll(boolean, java.lang.String[]) */
		public ProjOutput<V, U> projectAll(
				boolean maximum, String[] order) {
			/// @todo Auto-generated method stub
			assert false : "Not yet implemented";
			return null;
		}

		/** @see UtilitySolutionSpace#resolve() */
		public UtilitySolutionSpace<V, U> resolve() {
			/// @todo Auto-generated method stub
			assert false : "Not yet implemented";
			return null;
		}

		/** @see UtilitySolutionSpace#toHypercube() */
		public Hypercube<V, U> toHypercube() {
			/// @todo Auto-generated method stub
			assert false : "Not yet implemented";
			return null;
		}

		/** @see UtilitySolutionSpace#iteratorBestFirst(boolean, java.lang.String[], V[]) */
		public UtilitySolutionSpace.IteratorBestFirst<V, U> iteratorBestFirst(
				boolean maximize, String[] fixedVariables, V[] fixedValues) {
			/// @todo Auto-generated method stub
			assert false : "NotImplemented";
			return null;
		}

		/** @see UtilitySolutionSpace#projExpectMonotone(java.lang.String, java.util.Map, boolean) */
		public ch.epfl.lia.frodo.solutionSpaces.UtilitySolutionSpace.ProjOutput<V, U> projExpectMonotone(String varOut, Map< String, UtilitySolutionSpace<V, U> > distributions, boolean maximum) {
			/// @todo Auto-generated method stub
			assert false : "Not yet implemented";
			return null;
		}

		/** @see ch.epfl.lia.frodo.solutionSpaces.SolutionSpace#getOwner() */
		public String getOwner() {
			/// @todo Auto-generated method stub
			assert false : "Not yet implemented";
			return null;
		}

		/** @see ch.epfl.lia.frodo.solutionSpaces.SolutionSpace#setOwner(java.lang.String) */
		public void setOwner(String owner) {
			/// @todo Auto-generated method stub
			assert false : "Not yet implemented";
			
		}

		/** @see ch.epfl.lia.frodo.solutionSpaces.UtilitySolutionSpace#iterator(java.lang.String[], V[][], V[]) */
		public ch.epfl.lia.frodo.solutionSpaces.UtilitySolutionSpace.Iterator<V, U> iterator(
				String[] variables, V[][] domains, V[] assignment) {
			/// @todo Auto-generated method stub
			assert false : "Not yet implemented";
			return null;
		}
	}
	
	/** @see XCSPparser#getSpaces(Set, boolean, boolean, Set) */
	@Override
	protected List< ? extends UtilitySolutionSpace<V, U> > getSpaces (Set<String> vars, final boolean withAnonymVars, final boolean getProbs, 
			Set<String> forbiddenVars) {
		
		if (getProbs) 
			return super.getSpaces(vars, withAnonymVars, getProbs, forbiddenVars);
		
		List< ? extends UtilitySolutionSpace<V, U> > allSpaces = super.getSpaces(null, withAnonymVars, getProbs, forbiddenVars);
		
		List< UtilitySolutionSpace<V, U> > out = new ArrayList< UtilitySolutionSpace<V, U> > ();
		
		// First compute the backyard if necessary
		if (this.backyard == null) 
			this.computeBackyard(this.agentName);
		
		// Go through the list of backyard components to create each corresponding CompoundSpace
		int nbrComps = this.facadeConstraints.size();
		nextComp: for (int i = 0; i < nbrComps; i++) {
			
			// Skip this backyard component if its facade does not contain any of the input variables
			HashSet<String> facade = new HashSet<String> (this.facadeConstraints.get(i));
			if (vars != null && Collections.disjoint(vars, facade))
				continue;
			
			HashSet<String> backyardVars = this.backyardComponents.get(i);
			
			// Ditch this backyard component if its facade contains forbidden variables
			if (forbiddenVars != null) {
				for (String var2 : facade) {
					if (forbiddenVars.contains(var2)) {
						
						// Remove from allSpaces all constraints involving backyard variables in this backyard component
						nextSpace: for (Iterator< ? extends UtilitySolutionSpace<V, U > > iter = allSpaces.iterator(); iter.hasNext(); ) {
							for (String var3 : iter.next().getVariables()) {
								if (backyardVars.contains(var3)) {
									iter.remove();
									continue nextSpace;
								}
							}
						}
						
						continue nextComp;
					}
				}
			}
			
			// Remove anonymous variables from the facade if required
			if (! withAnonymVars) 
				for (Iterator<String> iter = facade.iterator(); iter.hasNext() ; ) 
					if (this.getOwner(iter.next()) == null) 
						iter.remove();
			
			if (backyardVars.isEmpty()) 
				continue;
			
			// Find all spaces involving at least one of the backyard variables in the current component
			ArrayList< UtilitySolutionSpace<V, U> > spaces = new ArrayList< UtilitySolutionSpace<V, U> > ();
			ext: for (Iterator< ? extends UtilitySolutionSpace<V, U> > iter = allSpaces.iterator(); iter.hasNext(); ) {
				UtilitySolutionSpace<V, U> space = iter.next();
				
				// Check if one of the variables in the space is a backyard variable of the current component
				for (String var2 : space.getVariables()) {
					if (backyardVars.contains(var2)) {
						
						// Record the space and move to the next one
						spaces.add(space);
						iter.remove();
						continue ext;
					}
				}
			}
			
			// Instantiate a corresponding CompoundSpace
			out.add(new CompoundSpace<V, U> (facade, backyardVars, spaces, this.maximize()));
		}
		
		// Finish with the remaining spaces that don't involve any backyard variables, unless they don't involve var
		if (vars == null) 
			out.addAll(allSpaces);
		else 
			for (UtilitySolutionSpace<V, U> space : allSpaces) 
				if (! Collections.disjoint(vars, Arrays.asList(space.getVariables()))) 
					out.add(space);
		
		return out;
	}
	
	/** @see XCSPparser#getOwners() */
	@Override
	public Map<String, String> getOwners () {
		
		// First compute the backyard if necessary
		if (this.backyard == null) 
			this.computeBackyard(this.agentName);
		
		Map<String, String> out = super.getOwners();
		
		// Remove any mention to backyard variables
		for (Iterator<String> iter = out.keySet().iterator(); iter.hasNext(); ) 
			if (this.backyard.contains(iter.next())) 
				iter.remove();
		
		return out;
	}
	
	/** @see XCSPparser#getNbrVars(String) */
	@SuppressWarnings("unchecked")
	@Override
	public int getNbrVars (String owner) {
		
		// First compute the backyard if necessary
		if (this.backyard == null) 
			this.computeBackyard(this.agentName);
		
		// Whether owner != agentName
		final boolean notMine = (super.agentName != null && ! super.agentName.equals(owner));
		
		int nbrVars = 0;
		for (Element var : (List<Element>) root.getChild("variables").getChildren()) 
			if (owner.equals(var.getAttributeValue("agent")) && (notMine || ! this.backyard.contains(var.getAttributeValue("name")))) 
				nbrVars++;
		
		return nbrVars;
	}
	
	/** @see XCSPparser#getVariables() */
	@SuppressWarnings("unchecked")
	@Override
	public Set<String> getVariables () {

		// First compute the backyard if necessary
		if (this.backyard == null) 
			this.computeBackyard(this.agentName);
		
		Set<String> out = new HashSet<String> ();

		for (Element varElmt : (List<Element>) root.getChild("variables").getChildren()) {
			String var = varElmt.getAttributeValue("name");
			if (varElmt.getAttributeValue("agent") != null && ! this.backyard.contains(var)) // ignore variables with no specified owner and backyard variables
				out.add(var);
		}

		return out;
	}
	
	/** @see XCSPparser#getVariables(String) */
	@SuppressWarnings("unchecked")
	@Override
	public Set<String> getVariables (String owner) {
		
		// First compute the backyard if necessary
		if (this.backyard == null) 
			this.computeBackyard(this.agentName);
		
		Set<String> out = new HashSet<String> ();

		if (owner == null) {
			for (Element var : (List<Element>) root.getChild("variables").getChildren()) 
				if (var.getAttributeValue("agent") == null) 
					out.add(var.getAttributeValue("name"));
			
		} else if (this.agentName == null || owner.equals(this.agentName)) {
			for (Element var : (List<Element>) root.getChild("variables").getChildren()) {
				String varName = var.getAttributeValue("name");
				if (owner.equals(var.getAttributeValue("agent")) && ! this.backyard.contains(varName)) 
					out.add(varName);
			}
		} else 
			for (Element var : (List<Element>) root.getChild("variables").getChildren()) 
				if (owner.equals(var.getAttributeValue("agent"))) 
					out.add(var.getAttributeValue("name"));

		return out;
	}
	
	/** @see XCSPparser#getNeighborVars(String, boolean) */
	@Override
	public HashSet<String> getNeighborVars (String var, final boolean withAnonymVars) {

		// First compute the backyard if necessary
		if (this.backyard == null) 
			this.computeBackyard(this.agentName);
		
		if (withAnonymVars) 
			return this.facadeNeighborhoods.get(var);
		
		// Remove anonymous variables
		HashSet<String> out = new HashSet<String> ();
		for (String neigh : this.facadeNeighborhoods.get(var)) 
			if (this.getOwner(neigh) != null) 
				out.add(neigh);
		
		return out;
	}

	/** @see XCSPparser#getNeighborhoods(String, boolean, boolean) */
	@Override
	public Map< String, HashSet<String> > getNeighborhoods (String agent, final boolean withAnonymVars, final boolean onlyAnonymVars) {

		// First compute the backyard if necessary
		if (this.backyard == null) 
			this.computeBackyard(this.agentName);
		
		if (! onlyAnonymVars && withAnonymVars) 
			return new HashMap< String, HashSet<String> > (this.facadeNeighborhoods);
		
		Map< String, HashSet<String> > out = new HashMap< String, HashSet<String> > ();
		
		for (Map.Entry< String, HashSet<String> > entry : this.facadeNeighborhoods.entrySet()) {
			
			// Skip this neighborhood if the variable is not owned by the input agent
			String var = entry.getKey();
			if (agent != null && ! agent.equals(this.getOwner(var))) 
				continue;
			
			HashSet<String> neighbors = new HashSet<String> ();
			for (String neigh : entry.getValue()) {
				if (onlyAnonymVars) {
					if (this.getOwner(neigh) == null) 
						neighbors.add(neigh);
				} else if (this.getOwner(neigh) != null) 
					neighbors.add(neigh);
			}
			
			out.put(var, neighbors);
		}
		
		return out;
	}
	
	/** @see XCSPparser#getBackyardNeighborhood(String) */
	@Override
	public Map< String, Collection<String> > getBackyardNeighborhood (String var) {

		// First compute the backyard if necessary
		if (this.backyard == null) 
			this.computeBackyard(this.agentName);
		
		/// @bug TBC
		assert false : "not implemented!";
		return null;
	}
	
}
