/*
FRODO: a FRamework for Open/Distributed Optimization
Copyright (C) 2008, 2009, 2010, 2011  Thomas Leaute, Brammert Ottens & Radoslaw Szymanek

FRODO is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FRODO is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.


How to contact the authors: 
<http://lia.epfl.ch/>

EPFL / IC / IIF / LIA
Batiment IN 
Station 14 
CH - 1015 Lausanne 
Switzerland
*/

package ch.epfl.lia.frodo.solutionSpaces.JaCoP;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.StringTokenizer;

import JaCoP.core.Store;
import JaCoP.constraints.*;
import JaCoP.core.IntDomain;
import JaCoP.core.IntVar;
import JaCoP.core.Var;

/**
 * Predicate wrapper that translates a predicate expression from XCSP into constraints
 * available in JaCoP. It accepts only functional representation. Possibly, 
 * some auxiliary variables will be created.
 * 
 * @author Radoslaw Szymanek and Krzysztof Kuchcinski for the initial version developed for JaCoP 2.4; 
 * @author Arnaud Jutzeler and Thomas Leaute for the modified version ported to FRODO and JaCoP 3. 
 */

public class Predicate extends DecomposedConstraint{

	/** The XCSP parameters of the constraint */
	String constraintParameters;

	/** Whether to print out debug information */
	final boolean debug = false;

	/** The functional representation of the predicate */
	String description;

	/** The constraint store */
	Store store;

	/** The XCSP parameters of the predicate */
	String predicateParameters;

	/** The auxiliary constraints created */
	ArrayList<Constraint> decompositionConstraints;
	
	/** The auxiliary variables created */
	ArrayList<Var> auxilaryVariables;
	
	/** The utility variable (if this is a soft constraint) */
	IntVar utilVar;
	
	/**
	 * It creates/imposes constraints into the store as soon as the Predicate
	 * constraint is being imposed.
	 * @param constraintParameters parameters to the constraint.
	 * @param predicateParameters parameters specified within a predicate definition.
	 * @param description description of the constraint specified as the predicate.
	 * @param store the constraint store in which context the constraints are being created/imposed.
	 */
	public Predicate(String constraintParameters, String predicateParameters,
					 String description, Store store) {

		this.store = store;

		this.constraintParameters = constraintParameters;
		this.predicateParameters = predicateParameters;
		this.description = description;

	}

	/** @see JaCoP.constraints.DecomposedConstraint#decompose(JaCoP.core.Store) */
	@Override
	public ArrayList<Constraint> decompose(Store store) {

		if (decompositionConstraints != null)
			return decompositionConstraints;
		
		decompositionConstraints = new ArrayList<Constraint>();
		auxilaryVariables = new ArrayList<Var>();
		
		StringTokenizer tokenizer = new StringTokenizer(predicateParameters, " ");
		StringTokenizer tokenizerConstraint = new StringTokenizer(
				constraintParameters, " ");

		HashMap<String, Object> variableMaping = new HashMap<String, Object>();

		while (tokenizer.hasMoreTokens()) {

			String nextToken = tokenizer.nextToken();

			if (nextToken.equals("int"))
				continue;
			
			String name = tokenizerConstraint.nextToken();
			
			Var temp = store.findVariable(name);
			
			if (temp == null)
				variableMaping.put(nextToken, Integer.valueOf(name));
			else
				variableMaping.put(nextToken, temp);

		}

		if (debug)
			System.out.println(variableMaping);

		/// TODO, test if trimming spaces does not cause problems.
		description = description.replace(" ", "");
		
		tokenizer = new StringTokenizer(description, "(,)");

		String nextToken = tokenizer.nextToken();

		Object token = parse(nextToken, tokenizer, store, variableMaping);
		assert token != null;

		if (debug)
			System.out.println(token);

		if (token instanceof Constraint) // hard constraint
			decompositionConstraints.add( (Constraint) token );
		
		else if (token instanceof IntVar) // soft constraint
			this.utilVar = (IntVar) token;
		
		else 
			System.err.println("Unrecognized token: " + token);

		return decompositionConstraints;
	}

	
    /**
     * It allows to obtain the constraint specified by the predicate
     * without imposing it.
     *
     * @param store the constraint store in which context the constraint is being created.
     * @return the constraint represented by this predicate constraint (expression).
     */

    public PrimitiveConstraint getConstraint(Store store) {

            StringTokenizer tokenizer = new StringTokenizer(predicateParameters,
                            " ");

            StringTokenizer tokenizerConstraint = new StringTokenizer(
                            constraintParameters, " ");

            HashMap<String, Object> variableMaping = new HashMap<String, Object>();

            while (tokenizer.hasMoreTokens()) {

                    String nextToken = tokenizer.nextToken();

                    if (nextToken.equals("int"))
                            continue;

                    String name = tokenizerConstraint.nextToken();

                    Var temp = store.findVariable(name);
                    if (temp == null)
                            variableMaping.put(nextToken, Integer.valueOf(name));
                    else
                            variableMaping.put(nextToken, temp);

            }

            if (debug)
                    System.out.println(variableMaping);

            description = description.replace(" ", "");
            
            tokenizer = new StringTokenizer(description, "(,)");

            String nextToken = tokenizer.nextToken();

            Object token = parse(nextToken, tokenizer, store, variableMaping);

            if (debug)
                    System.out.println(token);

            return (PrimitiveConstraint) token;

    }

	
	
	/**
	 * @param token 			the current token
	 * @param tokenizer 		the tokenizer
	 * @param store 			the constraint store
	 * @param variableMaping 	a mapping from name to variable or constant
	 * @return a constraint
	 */
	private Object parse(String token, StringTokenizer tokenizer, Store store,
			HashMap<String, Object> variableMaping) {
		
		try {
			return Integer.valueOf(token);
		} catch (Exception ex) {
			// if not an integer then just go on.
		}

        if (variableMaping.get(token) != null)
			return variableMaping.get(token);
		else {

			if (token.equals("abs")) {

				Object o1 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);

				if (o1 instanceof Integer) {
					String val = String.valueOf(o1);
					if (variableMaping.get(val) == null) 
						variableMaping.put(val, new IntVar(store, (Integer) o1, (Integer) o1));
					o1 = variableMaping.get(val);
				}

				if (o1 instanceof IntVar) {
					IntVar auxilary = new IntVar(store, IntDomain.MinInt,
												IntDomain.MaxInt);
					auxilaryVariables.add(auxilary);
					decompositionConstraints.add(new AbsXeqY((IntVar) o1, auxilary));
					return auxilary;
				}
			}
			
			if (token.equals("neg")) {

				Object o1 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);

				if (o1 instanceof Integer) {
					String val = String.valueOf(o1);
					if (variableMaping.get(val) == null) 
						variableMaping.put(val, new IntVar(store, (Integer) o1, (Integer) o1));
					o1 = variableMaping.get(val);
				}

				if (o1 instanceof IntVar) {
					IntVar auxilary = new IntVar(store, IntDomain.MinInt,
												IntDomain.MaxInt);
					auxilaryVariables.add(auxilary);
					decompositionConstraints.add(new XmulCeqZ((IntVar) o1, -1, auxilary));
					return auxilary;
				}
			}

			if (token.equals("sub")) {

				Object o1 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);
				Object o2 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);

				if (o1 instanceof IntVar && o2 instanceof IntVar) {
					IntVar auxilary = new IntVar(store, IntDomain.MinInt,
							IntDomain.MaxInt);
					auxilaryVariables.add(auxilary);
					decompositionConstraints.add(new XplusYeqZ((IntVar) o2, auxilary,
							(IntVar) o1));
					return auxilary;
				} else if (o1 instanceof IntVar && o2 instanceof Integer) {
					IntVar auxilary = new IntVar(store, IntDomain.MinInt,
							IntDomain.MaxInt);
					auxilaryVariables.add(auxilary);
					decompositionConstraints.add(new XplusCeqZ(auxilary, (Integer) o2,
							(IntVar) o1));
					return auxilary;
				} else if (o1 instanceof Integer && o2 instanceof IntVar) {
					IntVar auxilary = new IntVar(store, IntDomain.MinInt,
							IntDomain.MaxInt);
					auxilaryVariables.add(auxilary);
					decompositionConstraints.add(new XplusYeqC((IntVar) o2, auxilary,
							(Integer) o1));
					return auxilary;
				}

			}

			if (token.equals("add")) {

				Object o1 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);
				Object o2 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);

				if (o1 instanceof IntVar && o2 instanceof IntVar) {
					IntVar auxilary = new IntVar(store, IntDomain.MinInt,
							IntDomain.MaxInt);
					auxilaryVariables.add(auxilary);
					decompositionConstraints.add(new XplusYeqZ((IntVar) o1, (IntVar) o2,
							auxilary));
					return auxilary;
				} else if (o1 instanceof IntVar && o2 instanceof Integer) {
					IntVar auxilary = new IntVar(store, IntDomain.MinInt,
							IntDomain.MaxInt);
					auxilaryVariables.add(auxilary);
					decompositionConstraints.add(new XplusCeqZ((IntVar) o1, (Integer) o2,
							auxilary));
					return auxilary;
				} else if (o1 instanceof Integer && o2 instanceof IntVar) {
					IntVar auxilary = new IntVar(store, IntDomain.MinInt,
							IntDomain.MaxInt);
					auxilaryVariables.add(auxilary);
					decompositionConstraints.add(new XplusCeqZ((IntVar) o2, (Integer) o1,
							auxilary));
					return auxilary;
				}

			}

			if (token.equals("mul")) {

				Object o1 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);
				Object o2 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);

				if (o1 instanceof IntVar && o2 instanceof IntVar) {
					IntVar auxilary = new IntVar(store, IntDomain.MinInt,
							IntDomain.MaxInt);
					auxilaryVariables.add(auxilary);
					decompositionConstraints.add(new XmulYeqZ((IntVar) o1, (IntVar) o1,
							auxilary));
					return auxilary;
				} else if (o1 instanceof IntVar && o2 instanceof Integer) {
					IntVar auxilary = new IntVar(store, IntDomain.MinInt,
							IntDomain.MaxInt);
					auxilaryVariables.add(auxilary);
					decompositionConstraints.add(new XmulCeqZ((IntVar) o1, (Integer) o2,
							auxilary));
					return auxilary;
				} else if (o1 instanceof Integer && o2 instanceof IntVar) {
					IntVar auxilary = new IntVar(store, IntDomain.MinInt,
							IntDomain.MaxInt);
					auxilaryVariables.add(auxilary);
					decompositionConstraints.add(new XmulCeqZ((IntVar) o2, (Integer) o1,
							auxilary));
					return auxilary;
				}

			}

			if (token.equals("div")) {

				Object o1 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);
				Object o2 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);

				if (o1 instanceof IntVar && o2 instanceof IntVar) {
					IntVar T1 = new IntVar(store, IntDomain.MinInt, IntDomain.MaxInt);
					IntVar T2 = new IntVar(store, 0, IntDomain.MaxInt);
					IntVar T3 = new IntVar(store, 0, IntDomain.MaxInt);
					IntVar auxilary = new IntVar(store, IntDomain.MinInt, IntDomain.MaxInt);
					
					auxilaryVariables.add(T1);
					auxilaryVariables.add(T2);
					auxilaryVariables.add(T3);
					auxilaryVariables.add(auxilary);
					
					decompositionConstraints.add(new XmulYeqZ(auxilary, (IntVar) o2, T1));
					decompositionConstraints.add(new XplusYeqZ(T1, T2, (IntVar) o1));
					decompositionConstraints.add(new AbsXeqY((IntVar) o2, T3));
					decompositionConstraints.add(new XltY(T2, T3));
					
					return auxilary;
				} else if (o1 instanceof IntVar && o2 instanceof Integer) {
					IntVar T1 = new IntVar(store, IntDomain.MinInt, IntDomain.MaxInt);
					IntVar T2 = new IntVar(store, 0, IntDomain.MaxInt);
					IntVar auxilary = new IntVar(store, IntDomain.MinInt, IntDomain.MaxInt);
					
					auxilaryVariables.add(T1);
					auxilaryVariables.add(T2);
					auxilaryVariables.add(auxilary);
					
					decompositionConstraints.add(new XmulCeqZ(auxilary, (Integer) o2, T1));
					decompositionConstraints.add(new XplusYeqZ(T1, T2, (IntVar) o1));
					decompositionConstraints.add(new XltC(T2, Math.abs((Integer)o2)));
					
					return auxilary;
				} else if (o1 instanceof Integer && o2 instanceof IntVar) {
					IntVar T1 = new IntVar(store, IntDomain.MinInt, IntDomain.MaxInt);
					IntVar T2 = new IntVar(store, 0, IntDomain.MaxInt);
					IntVar T3 = new IntVar(store, 0, IntDomain.MaxInt);
					IntVar auxilary = new IntVar(store, IntDomain.MinInt, IntDomain.MaxInt);
					
					auxilaryVariables.add(T1);
					auxilaryVariables.add(T2);
					auxilaryVariables.add(T3);
					auxilaryVariables.add(auxilary);
					
					decompositionConstraints.add(new XmulYeqZ(auxilary, (IntVar) o2, T1));
					decompositionConstraints.add(new XplusYeqC(T1, T2, (Integer) o1));
					decompositionConstraints.add(new AbsXeqY((IntVar) o2, T3));
					decompositionConstraints.add(new XltY(T2, T3));
					
					return auxilary;
				}

			}

			if (token.equals("mod")) {

				Object o1 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);
				Object o2 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);

				if (o1 instanceof IntVar && o2 instanceof IntVar) {
					IntVar T1 = new IntVar(store, IntDomain.MinInt, IntDomain.MaxInt);
					IntVar T2 = new IntVar(store, IntDomain.MinInt, IntDomain.MaxInt);
					IntVar T3 = new IntVar(store, 0, IntDomain.MaxInt);
					IntVar auxilary = new IntVar(store, 0, IntDomain.MaxInt);
					
					auxilaryVariables.add(T1);
					auxilaryVariables.add(T2);
					auxilaryVariables.add(T3);
					auxilaryVariables.add(auxilary);
					
					decompositionConstraints.add(new XmulYeqZ(T1, (IntVar) o2, T2));
					decompositionConstraints.add(new XplusYeqZ(T2, auxilary, (IntVar) o1));
					decompositionConstraints.add(new AbsXeqY((IntVar) o2, T3));
					decompositionConstraints.add(new XltY(auxilary, T3));

					return auxilary;
				} else if (o1 instanceof IntVar && o2 instanceof Integer) {

					IntVar T1 = new IntVar(store, IntDomain.MinInt, IntDomain.MaxInt);
					IntVar T2 = new IntVar(store, IntDomain.MinInt, IntDomain.MaxInt);
					IntVar auxilary = new IntVar(store, 0, (Integer) o2 - 1);
					
					auxilaryVariables.add(T1);
					auxilaryVariables.add(T2);
					auxilaryVariables.add(auxilary);

					decompositionConstraints.add(new XmulCeqZ(T1, (Integer) o2, T2));
					decompositionConstraints.add(new XplusYeqZ(T2, auxilary, (IntVar) o1));
					decompositionConstraints.add(new XltC(auxilary, Math.abs((Integer)o2)));

					return auxilary;
				} else if (o1 instanceof Integer && o2 instanceof IntVar) {

					IntVar T1 = new IntVar(store, IntDomain.MinInt, IntDomain.MaxInt);
					IntVar T2 = new IntVar(store, IntDomain.MinInt, IntDomain.MaxInt);
					IntVar T3 = new IntVar(store, 0, IntDomain.MaxInt);
					IntVar auxilary = new IntVar(store, 0, IntDomain.MaxInt);
					
					auxilaryVariables.add(T1);
					auxilaryVariables.add(T2);
					auxilaryVariables.add(T3);
					auxilaryVariables.add(auxilary);
					
					decompositionConstraints.add(new XmulYeqZ(T1, (IntVar) o2, T2));
					decompositionConstraints.add(new XplusYeqC(T2, auxilary, (Integer) o1));
					decompositionConstraints.add(new AbsXeqY((IntVar) o2, T3));
					decompositionConstraints.add(new XltY(auxilary, T3));


					return auxilary;
				}
			}

			if (token.equals("pow")) {

				Object o1 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);
				Object o2 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);

				if (o1 instanceof Integer) {
					String val = String.valueOf(o1);
					if (variableMaping.get(val) == null)
						variableMaping.put(val,	new IntVar(store, (Integer) o1, (Integer) o1));
					o1 = variableMaping.get(val);
				}

				if (o2 instanceof Integer) {
					String val = String.valueOf(o2);
					if (variableMaping.get(val) == null)
						variableMaping.put(val,
								new IntVar(store, (Integer) o2, (Integer) o2));
					o2 = variableMaping.get(val);
				}
				

				IntVar auxilary = null;
				
				// If the variables can take negative values, we need to add specific variables and constraints as JaCoP native XexpYeqZ constraint
				// does not handle negative values.
				if(((IntVar) o1).dom().min() < 0){
					
					auxilary = new IntVar(store, IntDomain.MinInt,
							IntDomain.MaxInt);	
					auxilaryVariables.add(auxilary);
					
					IntVar v2 = new IntVar(store, 2, 2);
					auxilaryVariables.add(v2);
					
					IntVar reminder = new IntVar(store, 0, 1);
					auxilaryVariables.add(reminder);
					
					IntVar result = new IntVar(store, 0, IntDomain.MaxInt);
					auxilaryVariables.add(result);
					
					IntVar abs = new IntVar(store, 0, IntDomain.MaxInt);
					auxilaryVariables.add(abs);
					
					IntVar neg = new IntVar(store, IntDomain.MinInt, 0);
					auxilaryVariables.add(neg);
					
					decompositionConstraints.add(new AbsXeqY((IntVar) o1, abs));
					
					decompositionConstraints.add(new XmodYeqZ((IntVar) o2, v2, reminder));
					
					decompositionConstraints.add(new XmulCeqZ(result, -1, neg));
					
					if(((IntVar) o2).dom().min() < 0){
						
						IntVar abs2 = new IntVar(store, 0, IntDomain.MaxInt);
						auxilaryVariables.add(abs2);
						
						IntVar result2 = new IntVar(store, 0, IntDomain.MaxInt);
						auxilaryVariables.add(result2);
						
						decompositionConstraints.add(new AbsXeqY((IntVar) o2, abs2));
						
						decompositionConstraints.add(new XexpYeqZ(abs, abs2, result));
						
						decompositionConstraints.add(new IfThenElse(new XltC((IntVar) o1, 0), new IfThenElse(new XeqC(reminder, 0), new XeqY(result2, result), new XeqY(result2, neg)),  new XeqY(result2, result)));
						
						decompositionConstraints.add(new IfThenElse(new XltC((IntVar) o2, 0), new IfThenElse(new XeqC(abs, 1), new XeqY(auxilary, result2) , new XeqC(auxilary, -1)), new XeqY(auxilary, result2)));
						
					}else{
						
						decompositionConstraints.add(new XexpYeqZ(abs, (IntVar) o2, result));
						
						decompositionConstraints.add(new IfThenElse(new XltC((IntVar) o1, 0), new IfThenElse(new XeqC(reminder, 0), new XeqY(auxilary, result), new XeqY(auxilary, neg)), new XeqY(auxilary, result)));
					}
				}else{
					
					auxilary = new IntVar(store, 0,
							IntDomain.MaxInt);	
					auxilaryVariables.add(auxilary);
					
					if(((IntVar) o2).dom().min() < 0){
						
						IntVar result = new IntVar(store, 0, IntDomain.MaxInt);
						auxilaryVariables.add(result);
						
						IntVar abs = new IntVar(store, 0, IntDomain.MaxInt);
						auxilaryVariables.add(abs);
						
						decompositionConstraints.add(new AbsXeqY((IntVar) o2, abs));
						
						decompositionConstraints.add(new XexpYeqZ((IntVar) o1, (IntVar) abs, result));
						
						decompositionConstraints.add(new IfThenElse(new XltC((IntVar) o2, 0), new IfThenElse(new XeqC((IntVar) o1, 1), new XeqC(auxilary, 1), new XeqC(auxilary, -1)), new XeqY(auxilary, result)));
						
					// the domains only contain positive values
					}else{

						decompositionConstraints.add(new XexpYeqZ((IntVar) o1, (IntVar) o2, auxilary));
					}
				}
				
				///@bug when the value of o2 is 0,1 the JaCoP constraint XexpYeqZ shows invalid behaviour
				if(((IntVar) o1).dom().isIntersecting(-1, 1) ||
						((IntVar) o2).dom().isIntersecting(-1, 1)){
							System.err.println("The power operation is performed on variables that have -1, 0, 1 in their domains." +
									"\nThis can result in erroneous solutions due to some bugs in JaCoP.");
				}

				return auxilary;

			}
			
			if (token.equals("min")) {

				Object o1 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);
				Object o2 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);

				if (o1 instanceof IntVar && o2 instanceof IntVar) {
					IntVar auxilary = new IntVar(store, IntDomain.MinInt,
							IntDomain.MaxInt);
					auxilaryVariables.add(auxilary);
					
					IntVar[] listVars = {(IntVar) o1, (IntVar) o2};
					
					decompositionConstraints.add(new Min(listVars, auxilary));
					
					return auxilary;
				} else if (o1 instanceof IntVar && o2 instanceof Integer) {
					IntVar auxilary = new IntVar(store, IntDomain.MinInt,
							IntDomain.MaxInt);
					auxilaryVariables.add(auxilary);
					
					IntVar[] listVars = {(IntVar) o1, new IntVar(store, (Integer) o2, (Integer) o2)};
					
					decompositionConstraints.add(new Min(listVars, auxilary));
							
					return auxilary;
				} else if (o1 instanceof Integer && o2 instanceof IntVar) {
					IntVar auxilary = new IntVar(store, IntDomain.MinInt,
							IntDomain.MaxInt);
					auxilaryVariables.add(auxilary);
					
					IntVar[] listVars = {(IntVar) o2, new IntVar(store, (Integer) o1, (Integer) o1)};
					
					decompositionConstraints.add(new Min(listVars, auxilary));
				}

			}
			
			if (token.equals("max")) {

				Object o1 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);
				Object o2 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);

				if (o1 instanceof IntVar && o2 instanceof IntVar) {
					IntVar auxilary = new IntVar(store, IntDomain.MinInt,
							IntDomain.MaxInt);
					auxilaryVariables.add(auxilary);
					
					IntVar[] listVars = {(IntVar) o1, (IntVar) o2};
					
					decompositionConstraints.add(new Max(listVars, auxilary));
					
					return auxilary;
				} else if (o1 instanceof IntVar && o2 instanceof Integer) {
					IntVar auxilary = new IntVar(store, IntDomain.MinInt,
							IntDomain.MaxInt);
					auxilaryVariables.add(auxilary);
					
					IntVar[] listVars = {(IntVar) o1, new IntVar(store, (Integer) o2, (Integer) o2)};
					
					decompositionConstraints.add(new Max(listVars, auxilary));
							
					return auxilary;
				} else if (o1 instanceof Integer && o2 instanceof IntVar) {
					IntVar auxilary = new IntVar(store, IntDomain.MinInt,
							IntDomain.MaxInt);
					auxilaryVariables.add(auxilary);
					
					IntVar[] listVars = {(IntVar) o2, new IntVar(store, (Integer) o1, (Integer) o1)};
					
					decompositionConstraints.add(new Max(listVars, auxilary));
				}
			}
			
			if (token.equals("if")) {
				
				Object o1 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);
				Object o2 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);
				Object o3 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);

				if (o1 instanceof PrimitiveConstraint && o2 instanceof IntVar && o3 instanceof IntVar) {
					IntVar auxilary = new IntVar(store, IntDomain.MinInt,
							IntDomain.MaxInt);
					auxilaryVariables.add(auxilary);
					PrimitiveConstraint thenCons = new XeqY(auxilary, (IntVar) o2);
					PrimitiveConstraint elseCons = new XeqY(auxilary, (IntVar) o3);
					
					decompositionConstraints.add(new IfThenElse((PrimitiveConstraint) o1, 
							  thenCons,
							  elseCons));
					
					return auxilary;
				} else if (o1 instanceof PrimitiveConstraint && o2 instanceof IntVar && o3 instanceof Integer) {
					IntVar auxilary = new IntVar(store, IntDomain.MinInt,
							IntDomain.MaxInt);
					auxilaryVariables.add(auxilary);
					
					PrimitiveConstraint thenCons = new XeqY(auxilary, (IntVar) o2);
					PrimitiveConstraint elseCons = new XeqC(auxilary, (Integer) o3);
					
					decompositionConstraints.add(new IfThenElse((PrimitiveConstraint) o1, 
							  thenCons,
							  elseCons));
					
					return auxilary;
				} else if (o1 instanceof PrimitiveConstraint && o2 instanceof Integer && o3 instanceof IntVar) {
					IntVar auxilary = new IntVar(store, IntDomain.MinInt,
							IntDomain.MaxInt);
					auxilaryVariables.add(auxilary);
					
					PrimitiveConstraint thenCons = new XeqC(auxilary, (Integer) o2);
					PrimitiveConstraint elseCons = new XeqY(auxilary, (IntVar) o3);
					
					decompositionConstraints.add(new IfThenElse((PrimitiveConstraint) o1, 
							  thenCons,
							  elseCons));
					
					return auxilary;
				} else if (o1 instanceof PrimitiveConstraint && o2 instanceof Integer && o3 instanceof Integer) {
					IntVar auxilary = new IntVar(store, IntDomain.MinInt,
							IntDomain.MaxInt);
					auxilaryVariables.add(auxilary);
					
					PrimitiveConstraint thenCons = new XeqC(auxilary, (Integer) o2);
					PrimitiveConstraint elseCons = new XeqC(auxilary, (Integer) o3);
					
					decompositionConstraints.add(new IfThenElse((PrimitiveConstraint) o1, 
							  thenCons,
							  elseCons));
					
					return auxilary;
				}
				
				System.err.println("Failed to parse:\n\t if \n\t " + o1 + "\n\t then \n\t " + o2 + " \n\t else \n\t " + o3);
				return null;
			}

			if (token.equals("eq")) {
				
				String nextToken = tokenizer.nextToken();

				if (nextToken.equals("abs")) {
					
					String nextNextToken = tokenizer.nextToken();

					if (nextNextToken.equals("sub")) {

						Object o1 = parse(tokenizer.nextToken(), tokenizer,
								store, variableMaping);
						Object o2 = parse(tokenizer.nextToken(), tokenizer,
								store, variableMaping);
						Object o3 = parse(tokenizer.nextToken(), tokenizer,
								store, variableMaping);

						if (o1 instanceof Integer) {
							String val = String.valueOf(o1);
							if (variableMaping.get(val) == null)
								variableMaping.put(val, new IntVar(store, (Integer) o1,
										(Integer) o1));
							o1 = variableMaping.get(val);
						}

						if (o2 instanceof Integer) {
							String val = String.valueOf(o2);
							if (variableMaping.get(val) == null)
								variableMaping.put(val, new IntVar(store, (Integer) o2,
										(Integer) o2));
							o2 = variableMaping.get(val);
						}

						if (o3 instanceof Integer) {
							String val = String.valueOf(o3);
							if (variableMaping.get(val) == null)
								variableMaping.put(val, new IntVar(store, (Integer) o3,
										(Integer) o3));
							o3 = variableMaping.get(val);
						}

						return new Distance((IntVar) o1, (IntVar) o2,
								(IntVar) o3);

					} else {

						Object o1 = parse(nextNextToken, tokenizer, store,
								variableMaping);
						
						Object o2 = parse(tokenizer.nextToken(), tokenizer,
								store, variableMaping);

						if(o1 instanceof IntVar && o2 instanceof IntVar){
						
							return new AbsXeqY((IntVar) o1, (IntVar) o2);
						
						}else if(o1 instanceof IntVar && o2 instanceof Integer){
							
							IntVar auxilary = new IntVar(store, (Integer) o2,
									(Integer) o2);
							auxilaryVariables.add(auxilary);
							
							return new AbsXeqY((IntVar) o1, auxilary);
							
						}else if(o1 instanceof Integer && o2 instanceof IntVar){
							
							return new XeqC((IntVar) o2, Math.abs((Integer) o1));
							
						}
					}

				}

				if (nextToken.equals("add")) {
					
					Object o1 = parse(tokenizer.nextToken(), tokenizer, store,
							variableMaping);
					Object o2 = parse(tokenizer.nextToken(), tokenizer, store,
							variableMaping);
					Object o3 = parse(tokenizer.nextToken(), tokenizer, store,
							variableMaping);

					if (o1 instanceof IntVar && o2 instanceof IntVar
							&& o3 instanceof IntVar) {
						return (new XplusYeqZ((IntVar) o1, (IntVar) o2,
								(IntVar) o3));
					}

					if (o1 instanceof IntVar && o2 instanceof Integer
							&& o3 instanceof IntVar) {
						return (new XplusCeqZ((IntVar) o1, (Integer) o2,
								(IntVar) o3));
					}

					if (o1 instanceof Integer && o2 instanceof IntVar
							&& o3 instanceof IntVar) {
						return (new XplusCeqZ((IntVar) o2, (Integer) o1,
								(IntVar) o3));
					}

					if (o1 instanceof IntVar && o2 instanceof IntVar
							&& o3 instanceof Integer) {
						return (new XplusYeqC((IntVar) o1, (IntVar) o2,
								(Integer) o3));
					}

				}

				if (nextToken.equals("mul")) {

					Object o1 = parse(tokenizer.nextToken(), tokenizer, store,
							variableMaping);
					Object o2 = parse(tokenizer.nextToken(), tokenizer, store,
							variableMaping);
					Object o3 = parse(tokenizer.nextToken(), tokenizer, store,
							variableMaping);

					if (o1 instanceof IntVar && o2 instanceof IntVar
							&& o3 instanceof IntVar) {
						return (new XmulYeqZ((IntVar) o1, (IntVar) o2,
								(IntVar) o3));
					}

					if (o1 instanceof IntVar && o2 instanceof Integer
							&& o3 instanceof Var) {
						return (new XmulCeqZ((IntVar) o1, (Integer) o2,
								(IntVar) o3));
					}

					if (o1 instanceof Integer && o2 instanceof IntVar
							&& o3 instanceof Var) {
						return (new XmulCeqZ((IntVar) o2, (Integer) o1,
								(IntVar) o3));
					}

					if (o1 instanceof IntVar && o2 instanceof IntVar
							&& o3 instanceof Integer) {
						return (new XmulYeqC((IntVar) o1, (IntVar) o2,
								(Integer) o3));
					}

				}

				// System.out.println("nextToken " + nextToken);

				Object o1 = parse(nextToken, tokenizer, store, variableMaping);
				Object o2 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);

				// System.out.println(o1);
				// System.out.println(o2);

				if (o1 instanceof IntVar && o2 instanceof IntVar)
					return new XeqY((IntVar) o1, (IntVar) o2);
				else if (o1 instanceof IntVar && o2 instanceof Integer)
					return new XeqC((IntVar) o1, (Integer) o2);
				else if (o1 instanceof Integer && o2 instanceof Var)
					return new XeqC((IntVar) o2, (Integer) o1);
				else
					return new Eq((PrimitiveConstraint) (o1),
							(PrimitiveConstraint) (o2));
			}

			if (token.equals("ne")) {

				Object o1 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);
				Object o2 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);

				if (o1 instanceof IntVar && o2 instanceof IntVar)
					return new XneqY((IntVar) o1, (IntVar) o2);
				else if (o1 instanceof IntVar && o2 instanceof Integer)
					return new XneqC((IntVar) o1, (Integer) o2);
				else if (o1 instanceof Integer && o2 instanceof IntVar)
					return new XneqC((IntVar) o2, (Integer) o1);
				else
					return new Eq(new Not((PrimitiveConstraint) (o1)),
							(PrimitiveConstraint) (o2));
			}

			if (token.equals("ge")) {
				Object o1 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);
				Object o2 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);

				if (o1 instanceof IntVar && o2 instanceof IntVar)
					return new XgteqY((IntVar) o1, (IntVar) o2);
				else if (o1 instanceof IntVar && o2 instanceof Integer)
					return new XgteqC((IntVar) o1, (Integer) o2);
				else if (o1 instanceof Integer && o2 instanceof IntVar)
					return new XlteqC((IntVar) o2, (Integer) o1);
				else
					System.out.println("ge predicate does not have variable parameters");
			}

			if (token.equals("gt")) {
				Object o1 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);
				Object o2 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);

				if (o1 instanceof IntVar && o2 instanceof IntVar)
					return new XgtY((IntVar) o1, (IntVar) o2);
				else if (o1 instanceof IntVar && o2 instanceof Integer)
					return new XgtC((IntVar) o1, (Integer) o2);
				else if (o1 instanceof Integer && o2 instanceof IntVar)
					return new XltC((IntVar) o2, (Integer) o1);
				else
					System.out.println("gt predicate does not have variable parameters");
			}

			if (token.equals("le")) {

				String nextToken = tokenizer.nextToken();

				if (nextToken.equals("add")) {

					Object o1 = parse(tokenizer.nextToken(), tokenizer, store,
							variableMaping);
					Object o2 = parse(tokenizer.nextToken(), tokenizer, store,
							variableMaping);
					Object o3 = parse(tokenizer.nextToken(), tokenizer, store,
							variableMaping);

					if (o1 instanceof IntVar && o2 instanceof IntVar
							&& o3 instanceof IntVar) {
						return (new XplusYlteqZ((IntVar) o1, (IntVar) o2,
								(IntVar) o3));
					}

					if (o1 instanceof IntVar && o2 instanceof Integer
							&& o3 instanceof IntVar) {
						return (new XplusClteqZ((IntVar) o1, (Integer) o2,
								(IntVar) o3));
					}

					if (o1 instanceof Integer && o2 instanceof IntVar
							&& o3 instanceof IntVar) {
						return (new XplusClteqZ((IntVar) o2, (Integer) o1,
								(IntVar) o3));
					}

					if (o1 instanceof IntVar && o2 instanceof IntVar
							&& o3 instanceof Integer) {

						String val = String.valueOf(o3);
						if (variableMaping.get(val) == null)
							variableMaping.put(val, new IntVar(store, (Integer) o3,
									(Integer) o3));
						o3 = variableMaping.get(val);

						return (new XplusYlteqZ((IntVar) o1, (IntVar) o2,
								(IntVar) o3));
					}

				}

				Object o1 = parse(nextToken, tokenizer, store, variableMaping);
				Object o2 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);

				if (o1 instanceof IntVar && o2 instanceof IntVar)
					return new XlteqY((IntVar) o1, (IntVar) o2);
				else if (o1 instanceof IntVar && o2 instanceof Integer)
					return new XlteqC((IntVar) o1, (Integer) o2);
				else if (o1 instanceof Integer && o2 instanceof IntVar)
					return new XgteqC((IntVar) o2, (Integer) o1);
				else
					System.out.println("le predicate does not have variable parameters");
			}

			if (token.equals("lt")) {
				Object o1 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);
				Object o2 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);

				if (o1 instanceof IntVar && o2 instanceof IntVar)
					return new XltY((IntVar) o1, (IntVar) o2);
				else if (o1 instanceof IntVar && o2 instanceof Integer)
					return new XltC((IntVar) o1, (Integer) o2);
				else if (o1 instanceof Integer && o2 instanceof IntVar)
					return new XgtC((IntVar) o2, (Integer) o1);
				else
					System.out.println("lt predicate does not have variable parameters");
			}

			if (token.equals("not")) {
				Object o1 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);

				return new Not((PrimitiveConstraint) o1);
			}

			if (token.equals("and")) {
				Object o1 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);
				Object o2 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);

				return new And((PrimitiveConstraint) o1,
						(PrimitiveConstraint) o2);
			}

			if (token.equals("or")) {
				Object o1 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);
				Object o2 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);

				return new Or((PrimitiveConstraint) o1,
						(PrimitiveConstraint) o2);
			}

			if (token.equals("xor")) {
				Object o1 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);
				Object o2 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);

				return new Eq(new Not((PrimitiveConstraint) (o1)),
						(PrimitiveConstraint) (o2));
			}
			
			if (token.equals("iff")) {
				Object o1 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);
				Object o2 = parse(tokenizer.nextToken(), tokenizer, store,
						variableMaping);

				return new Eq((PrimitiveConstraint) (o1),
						(PrimitiveConstraint) (o2));
			}

			System.err.println("Unknown token: `" + token + "'");
			return null;

		}

	}


	/** @see JaCoP.constraints.DecomposedConstraint#imposeDecomposition(JaCoP.core.Store) */
	@Override
	public void imposeDecomposition(Store store) {
		
		if (decompositionConstraints == null)
			decompose(store);
		
		for (Constraint c : decompositionConstraints)
			store.impose(c);
		
		store.auxilaryVariables.addAll(auxilaryVariables);
	}

	/** @see JaCoP.constraints.DecomposedConstraint#auxiliaryVariables() */
	@Override
	public ArrayList<Var> auxiliaryVariables() { 
		return auxilaryVariables;
	}
	
}
